import json
from bffacilities._constants import _CONFIG_DIR
from bffacilities import createLogger
import os
import os.path as osp

logger = createLogger("checkGitRepo", savefile=False, stream=True)

os.makedirs(_CONFIG_DIR, exist_ok=True)
configFile = _CONFIG_DIR / "checkGitRepo.json"

def addDirIntoMonitor(dirs):
    """dirs: array(abspath)
    """
    content = None
    try :
        with open(configFile, "r") as f:
            fileContent = f.read()
            if len(fileContent) > 0:
                content = json.loads(fileContent)
    except Exception as e:
        logger.error(f"error reading configFile {configFile}: {e}")

    if content is None:
        content = []
    added = 0
    for d in dirs:
        if d in content:
            logger.warning(f"Directory {d} already added.")
            continue
        added += 1
        logger.info(f"Directory {d} will be added.")
        content.append(d)

    if added:
        try:
            with open(configFile, "w") as f:
                json.dump(content, f)
        except Exception as e:
            logger.error(f"error writing configFile {configFile}: {e}")
def delDirFromMonitor(dirs):
    content = None
    try :
        with open(configFile, "r") as f:
            content = json.load(f)
    except Exception as e:
        logger.error(f"error reading configFile {configFile}: {e}")

    if content is None:
        logger.warning(f"Directory does not exist in empty monitor list.")
        return
    deled = 0
    for d in dirs:
        if d not in content:
            logger.warning(f"Directory does not exist in empty monitor list.")
            continue
        logger.info(f"directory {d} will be removed.")
        deled += 1
        content.remove(d)
    with open(configFile, "w") as f:
        json.dump(content, f)

import subprocess as sp
import io
def executeGit(cmd):
    content = None
    try :
        with open(configFile, "r") as f:
            content = json.load(f)
    except Exception as e:
        logger.error(f"Error reading configFile {configFile}: {e}")
    if content is None:
        logger.error(f"No directory added in monitor list, please add them first.")
        return
    assert type(content) is list, "Error content"
    git = ["git"]
    git.extend(cmd)

    # d = content[0]
    processes = []
    for d in content:
        if not osp.exists(d):
            logger.warning(f"Directory {d} already deleted, consider to remove it from monitor list")
            continue
        logger.info(f"================ Entering directory {d} ===============\n")
        p = sp.Popen(git, cwd=d)
        p.wait()
        print("\n")
        # processes.append((d, p))
        # print(p.stdout.read())
    # strIOs = []
    # for d, p in processes:
    #     p.wait()
    #     strIO = io.TextIOWrapper(io.BytesIO(p.stdout.read()))
    #     strIOs.append((d, strIO))
    # for d, strIO in strIOs:
    #     logger.info(f"================ Entering directory {d} ===============\n")
    #     print(strIO.read())
def validateDirs(directories):
    dirs = []
    for d in directories:
        d = osp.abspath(d)
        if not osp.exists(osp.join(d, ".git")):
            logger.warning(f"Directory {d} may not be git repository.")
            continue
        if osp.exists(d):
            dirs.append(d)
    return dirs
def main(args = None):
    from argparse import ArgumentParser
    parser = ArgumentParser(prog="checkGitRepo", description="Check git repo and do some actions")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-a", "--add-dir", type=str, nargs="+", help="The directory will be added into program monitor list")
    group.add_argument("-d", "--del-dir", type=str, nargs="+", help="Delete directory from monitor list")
    group.add_argument("-e", "--execute", type=str, nargs="+", help="execute git command, for example: checkGitRepo -e 'reset -- .' will call 'git reset -- .'")
    group.add_argument("-s", "--status", action="store_true", help="equvalent of 'git status'")
    group.add_argument("-o", "--open", action="store_true", help="Open config file")

    args = vars(parser.parse_args(args))

    if args["add_dir"]:
        directories = args["add_dir"]
        dirs = validateDirs(directories)
        addDirIntoMonitor(dirs)
            # print(d)
    elif args["del_dir"]:
        directories = args["del_dir"]
        dirs = []
        for d in directories:
            dirs.append(osp.abspath(d))
        delDirFromMonitor(dirs)
    elif args["status"]:
        executeGit(["status"])
    elif args["execute"]:
        executeGit(args["execute"])
    elif args["open"]:
        import webbrowser
        webbrowser.open(configFile)
    else:
        parser.print_help()





import collections
import datetime
import glob
import uuid
import sys
import numpy as np
import labelme
import re
SubCategoryPatter = re.compile("(\w+)-(\w+|\d+)")

try:
    import pycocotools.mask as cocomask
except ImportError:
    print("Please install pycocotools:\n\n    pip install pycocotools\n")
    sys.exit(1)

class Labelme2Cocoer():
    """

    """
    def __init__(self, output_dir, cocomask=None, debug=False):
        """
        Arguments:
            output_dir(str)  annotations.json and image will store in this place
            
            cocomask(pytcocotools.mask): 
                @deprecated
        """
        self.output_dir = output_dir
        self.debug = debug
        os.makedirs(output_dir, exist_ok=True)
        os.makedirs(osp.join(output_dir, "JPEGImages"), exist_ok=True)
        now = datetime.datetime.now()

        ## coco data that will be write into annotations.json
        self._data_info = dict(
                description=None,
                url=None,
                version=None,
                year=now.year,
                contributor=None,
                date_created=now.strftime("%Y-%m-%d %H:%M:%S.%f"),
            )
        self._data_licenses= [dict(url=None, id=0, name=None,)]
        self._data_images=[
                # license, url, file_name, height, width, date_captured, id
            ]
            # supercategory, id, name
        self._data_annotations=[
                # segmentation, area, iscrowd, image_id, bbox, category_id, id
            ]
        self._data_categories= None

    def classNameToId(self, labels_file, skeletondict=None):
        """
        @deprecated see setClassId

        读取 labels 文件，将 className 转换为 id，
        对于关键点名称，需要包含 Mask 名称作为前缀，并用 - 分隔

        对于关键点来说，必须有父类才行

        Currently keypoint labelname should be `xxx-1` or `xxx-2`, because it's
        easy to be ordered.

        Argument:
            labels_file (str): labels file loc,
        """
        class_id = -1  # class id 仅会给对象分配
        # dict to map class_name (str) to id (int)
        class_name_to_id = {} # key: class_name， 仅包含对象名, value: id
        categoriedict = {}
        with open(labels_file, "r") as f:
            for i, line in enumerate(f.readlines()):
                class_name = line.strip()   
                # 形如 class-subclass1
                if len(class_name) == 0:
                    continue
                
                if class_id == -1:
                    assert class_name == "__ignore__"
                    class_id += 1
                    continue

                matched = SubCategoryPatter.match(class_name)
                class_name_to_id[class_name] = class_id
                if matched is None:
                    # 检测为对象
                    categoriedict[class_id] = { 
                        "name": class_name, 
                        "supercategory": None,
                        "type": "object", 
                        "keypoints": [],
                    }
                    class_id += 1
                else:
                    # 检测为关键点
                    kp_name = class_name
                    class_name = matched[1]

                    class_id = class_name_to_id[class_name]
                    # keypoint name 也映射到相同的 id 
                    class_name_to_id[kp_name] = class_id

                    cat = categoriedict[class_id]
                    cat["type"] = "keypoint"
                    cat["keypoints"].append(kp_name)
        if self.debug:
            print("Class to id", class_name_to_id, "============")
            # raise ValueError("Test")
        self.class_name_to_id = class_name_to_id

        if self.debug:
            print("Categories Dict", categoriedict)

        categories = []
        skeleton = [] if skeletondict is None else skeletondict
        for class_id, cat in categoriedict.items():
            if cat["type"] == "object":
                categories.append({
                    "supercategory": cat["supercategory"], 
                    "id": class_id, 
                    "name": cat["name"],
                })
            elif cat["type"] == "keypoint":
                categories.append({
                    "supercategory": cat["supercategory"], 
                    "id": class_id, 
                    "name": cat["name"],
                    "keypoints": cat["keypoints"],
                    "skeleton": skeleton
                })

        self._data_categories = list(sorted(categories, key=lambda x: x["id"]))

        if self.debug:
            print("cats", self._data_categories, "============\n\n")

    def setClassId(self, class_dict):
        """ 默认背景 classId 为 0
        Args
            class_dict
            {
                "background": True, // default
                "instances": ["L1_1", "L1_2", "L1_3],
                // L1_1 L1_2 分别有关键点, L1_3 没有
                "keypoints": {
                    "L1_1": ["L2_1", "L2_2"]
                    "L1_2": ["L2_3"]
                },
                "skeleton": {

                }
            }
        """

        background = class_dict.get("background", True)
        categories = []
        if background:
            categories.append({
                "supercategory": None,
                "id": 0,
                "name": "_background_"
            })
        class_name_to_id = {
            "_background_": 0,
        }

        instances = class_dict["instances"]
        keypoints = class_dict["keypoints"] if "keypoints" in class_dict else {}
        skeleton = class_dict["skeleton"] if "skeleton" in class_dict else {}

        for i, ins_name in enumerate(instances):
            kps = keypoints.get(ins_name, [])
            skt = skeleton.get(ins_name, [])
            categories.append({
                "supercategory": None,
                "id": i + 1,
                "name": ins_name,
                "keypoints": kps,
                "skeleton": skt
            })
            class_name_to_id[ins_name] = i + 1
        self._data_categories = categories

        self.class_name_to_id = class_name_to_id
        keymap = {}
        for obj, kps in keypoints.items():
            for kp in kps:
                keymap[kp] = obj
        self._class_dict = class_dict
        self._class_dict["keymap"] = keymap

    def _kps_from_categories(self, cls_name):
        for categories in self._data_categories:
            if categories["name"] == cls_name:
                return categories["keypoints"]

    def _extractAnno(self, label_file, imshape):
        """
        this function can only find keypoints for each segclass on one image,
        
        if multiple segmentation class is on one image, group id must be set in labelme

        one part of the keypoints of segmentations will be saved

        Returns
            masks: key 为 (label, group_id)，value 为 mask

            keypoints: key 为 object class name，value 为 list((label_name, points))
        """
        masks = {}
        segmentations = collections.defaultdict(list)  # for segmentation
        keypoints = {}

        for shape in label_file.shapes:
            shape_type = shape.get("shape_type", "polygon")

            points = shape["points"]
            label = shape["label"]
            group_id = shape.get("group_id")

            if shape_type == "point": 
                # keypoint  is special
                # matched = SubCategoryPatter.match(label)
                segclass_name = self._class_dict["keymap"][label]
                if segclass_name not in keypoints:
                    keypoints[segclass_name] = []
                # points[0] 即该关键点
                keypoints[segclass_name].append((label, points[0]))
                continue

            mask = labelme.utils.shape_to_mask(imshape, points, shape_type)
            if group_id is None:
                # 保证在一副图片上的对象都是不同的 group_id
                group_id = uuid.uuid1()

            instance = (label, group_id)

            # if some instance is the same, then their masks should be intersected
            if instance in masks:
                masks[instance] = masks[instance] | mask
            else:
                masks[instance] = mask

            if shape_type == "rectangle":
                (x1, y1), (x2, y2) = points
                x1, x2 = sorted([x1, x2])
                y1, y2 = sorted([y1, y2])
                # clock wise
                points = [x1, y1, x2, y1, x2, y2, x1, y2]
            else:
                # 这里直接展平，不考虑分割成几部分的情况
                points = np.asarray(points).flatten().tolist()

            segmentations[instance].append(points)
        return masks, segmentations, keypoints

    def _convert_kp(self, cls_name, keypoints, imshape=None):
        kplist = []
        kpnum = 0
        i = 0
        keypoint_in_category = self._kps_from_categories(cls_name)
        boundary = [10000, 100000, 0, 0] # Top Left  Bottom Right
        if cls_name in keypoints:
            for kp in keypoints[cls_name]:
                # kp is the 2-element-tuple list
                name = kp[0]
                if name not in keypoint_in_category:
                    raise ValueError("KeyPoint and class not match")
                assert i == keypoint_in_category.index(name)
                i += 1
                point = kp[1]
                x = point[0]
                y = point[1]
                kplist.append(x) # x
                kplist.append(y) # y
                kplist.append(2)
                kpnum += 1

                if x < boundary[1]: boundary[1] = x
                if x > boundary[3]: boundary[3] = x
                if y < boundary[0]: boundary[0] = y
                if y > boundary[2]: boundary[2] = y
        if imshape is not None:
            boundary[0] -= 50; boundary[1] -= 50
            boundary[2] += 50; boundary[3] += 50
            if boundary[0] < 0: boundary[0] = 0
            if boundary[1] < 0: boundary[1] = 0
            if boundary[2] > imshape[0]: boundary[2] = imshape[0]
            if boundary[3] > imshape[1]: boundary[3] = imshape[1]
        return kplist, kpnum, boundary
           
    def _getAnno(self, label_file, img, image_id):
        """
        Extract all annotations from single labeled json file.

        Arguments:
            label_file: label_file object
            img: img_arr
            image_id (int): 
        """
        # masks is for area
        masks, segmentations, keypoints = self._extractAnno(label_file, img.shape[:2])
        segmentations = dict(segmentations)
        # sort keypoints to get ordered value
        for _, kp in keypoints.items():
            kp.sort(key=lambda x: x[0])

        if self.debug:
            print("Mask: ", len(masks), masks.keys())
            print("Segm", len(segmentations))
            # for segclass, items in keypoints.items():
            #     print("Keypoints: ", segclass, items)
            self.masks = masks

        annotations = self._data_annotations
        if len(masks) > 0:
            for instance, mask in masks.items():
                cls_name, group_id = instance
                if cls_name not in self.class_name_to_id:
                    continue
                cls_id = self.class_name_to_id[cls_name]

                mask = np.asfortranarray(mask.astype(np.uint8))
                mask = cocomask.encode(mask)
                bbox = cocomask.toBbox(mask).flatten().tolist()
                # it calculates mask area, not box area that torch needs
                area = float(cocomask.area(mask)) 
                # following is the area that torch needs
                # area = float(bbox[2] * bbox[3])
                # bbox = [bbox[0], bbox[1], bbox[0], bbox[1]] # format for torch
                kplist, kpnum, _ = self._convert_kp(cls_name, keypoints)
                anno = dict(
                    id=len(annotations),
                    image_id=image_id,
                    category_id=cls_id,
                    segmentation=segmentations[instance],
                    area=area, 
                    bbox=bbox, 
                    iscrowd=0,
                    keypoints = kplist, # [[kp[0], kp[1], 1] for kp in kps ], # format for torch
                    num_keypoints = kpnum,
                )
                annotations.append(anno)
        else:
            # 这种情况对应只有关键点，没有 mask 的情况
            assert len(keypoints) == 1, "Unable to extract annotation from this file"
            # print("No mask found\n\n")
            # if cls_name not in keypoints:
                # print("Error in class name and keypoint list")
                
            cls_name = list(keypoints.keys())[0] # 找到 keypoint 对应的类名，这里默认只有一个类
            cls_id = self.class_name_to_id[cls_name]
   
            kplist = []
            kpnum = 0
            # print(kps, "========")
            kplist, kpnum, boundary = self._convert_kp(cls_name=cls_name, keypoints=keypoints, imshape=img.shape)
            bbox = [boundary[1], boundary[0], (boundary[3] - boundary[1]), (boundary[2] - boundary[0])]
            area = (boundary[2] - boundary[0]) * (boundary[3] - boundary[1])
            segmentation = [ 
                boundary[1], boundary[0], 
                boundary[1], boundary[2], 
                boundary[3], boundary[2], 
                boundary[3], boundary[0], 
            ]
            anno = dict(
                id=len(annotations),
                image_id=image_id,
                category_id=cls_id,
                segmentation=[segmentation],
                area=area, 
                bbox=bbox, 
                iscrowd=0,
                keypoints = kplist, # [[kp[0], kp[1], 1] for kp in kps ], # format for torch
                num_keypoints = kpnum,
            )
            annotations.append(anno)
        if self.debug:
            print(annotations)

    def generateCocoJson(self, label_files):
        """
        生成 COCO 格式的 JSON 文件

        Args
            label_files: iterateble filenames
        """
        
        out_ann_file = osp.join(self.output_dir, "annotations.json")
        for image_id, filename in tqdm(enumerate(label_files)):
            try:
                label_file = labelme.LabelFile(filename=filename)
            except:
                print("Error generating dataset from:", filename)
                continue
            # 
            base = osp.splitext(osp.basename(filename))[0]
            out_img_file = osp.join(self.output_dir, "JPEGImages", base + ".jpg")

            img = labelme.utils.img_data_to_arr(label_file.imageData)
            Image.fromarray(img).convert("RGB").save(out_img_file)
            ## 填充 images
            self._data_images.append(
                dict(
                    license=0,
                    url=None,
                    file_name=osp.relpath(out_img_file, osp.dirname(out_ann_file)),
                    height=img.shape[0],
                    width=img.shape[1],
                    date_captured=None,
                    id=image_id,
                )
            )

            self._getAnno(label_file, img, image_id)

            if self.debug:
                break


    def output(self, out_ann_file = None, dtype="instances"):
        if out_ann_file is None:
            out_ann_file = "annotations.json"
        out_ann_file = osp.join(self.output_dir, out_ann_file)
        with open(out_ann_file, "w") as f:
            data = dict(
                type=dtype,
                annotations=self._data_annotations,
                categories=self._data_categories,
                images=self._data_images,
                info=self._data_info,
                licenses=self._data_licenses
            )
            json.dump(data, f, indent=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Description: Monitor folder's change and restart predefined Program, Original File is copied from Liaoxuefei's tutorial.
Monitor file changes, and execute prepared commands.
I made some modifications and optimizations. Now the changed filename will be send to the program and you can use {name} within
cmd arguments to get the name.

Author: BriFuture
"""

__author__ = 'BriFuture'
__version__ = '0.2.1'

import os
import sys

from bffacilities.utils import createLogger, initGetText
from pathlib import Path
import json
from argparse import ArgumentParser
from bffacilities._frame import Frame

tr = initGetText("monitor")
logger = createLogger("monitor.log", stream=True)

class Configuration(Frame):
    _DEFAULT_LOC = Path(os.getcwd()) / ".bfc" 
    _DEFAULT_FILE = Path(os.getcwd()) / ".bfc" / "monitor.json"

    def __init__(self):
        self.config = {
            # command name, str
            "cmd": None,
            # command arguments,  []
            "cmd_args": None,
            # monitor directory, "."
            "mon_dir": ".",

            "recursive": True,
            # monitor file extensions, for example: [".py", ".js"]
            "mon_ext": None,
            # exclusion file names, file extension needed, for example "test.py"
            "exclude": None
        }
        self._addArgs()

    def readConfig(self, file: Path):
        if not file or not file.exists():
            return
        with file.open("r", encoding="utf-8") as f:
            self.config = json.loads(f.read())
            # wont update cmdline args
            # for k, v in config.items():
            #     if k not in self.config:
            #         self.config[k] = v

    def _addArgs(self):
        parser = ArgumentParser(description=tr("Monitor file changes, and execute prepared commands."))

        self.parser = parser
        argument = parser.add_mutually_exclusive_group(required=False)
        argument.add_argument("-c", "--cmd", help=tr("""Specify cmd to execute after file events triggered. If the command have its options,
            for example 'python -V', use quote \" to wrap the whole command. It's recommend on parsing arguments in this way, so you can use {{name}}
            to identify which file has been changed"""))
        argument.add_argument("-C", "--config", help=tr("Read from config file, only file name is needed, for example, if you type '--config test', then this script will find a file named test.json which locates under {}. Note: The contents in the file will override other options given as command arguments.")
            .format(Configuration._DEFAULT_FILE))

        parser.add_argument("-a", "--argument", nargs="*", help=tr("Append command arguments for specified CMD. These arguments should not start with '-'"))
        parser.add_argument("-d", "--directory", help=tr("The directory to monitor, . by default."))
        parser.add_argument("-r", "--recursive", help=tr("Recusively monitor the diretory and sub directories"), action="store_true")
        parser.add_argument("-e", "--extension", nargs="*", type=str, help=tr("Specify the file suffix that needs to monitor, .py extension by default"))
        parser.add_argument("-E", "--exclude", nargs="*", type=str, help=tr("File that should be excluded when events triggerd"))
        parser.add_argument("-s", "--save-config", help=tr("Save the cofiguration into a file, the default file is {}").format(self._DEFAULT_FILE), nargs="?")
        parser.add_argument("-S", "--start-at-first", help=tr("Start the process or command when this scripts running."), action="store_true")
        parser.add_argument("--version", help=tr("Print version and exit."), action="version", version="Monitor's version is {}".format(__version__))


    def parseArgs(self, argv):
        args = self.parser.parse_args(argv)

        if args.config is None and args.cmd is None:
            logger.critical("Neither a config file is specified, nor cmd is given.")
            sys.exit(1)

        if args.config:
            self.readConfig(args.config)
            return

        if args.cmd is not None and " " in args.cmd:
            self.config["cmd"] = args.cmd.split(" ")
        else:
            self.config["cmd"] = [args.cmd]
        self.config["cmd_args"] = args.argument or []
        # , default is [".py"]
        self.config["mon_ext"] = args.extension or [".py"]
        mon_dir = args.directory or os.getcwd()
        self.config["mon_dir"] = str(Path(mon_dir).resolve())
        self.config["exclude"] = args.exclude or []
        self.config["recursive"] = args.recursive or True
        self.config["start"] = args.start_at_first

        if args.save_config:
            sc = self._DEFAULT_FILE / "monitor_{}.json".format(args.save_config)
        else:
            sc = self._DEFAULT_FILE 

        with sc.open("w", encoding="utf-8") as f:
            # mon_dir = self.config["mon_dir"]
            # self.config["mon_dir"] = str(mon_dir)
            json.dump(self.config, f)
            # self.config["mon_dir"] = mon_dir
Configuration._DEFAULT_LOC.mkdir(parents=True, exist_ok=True)

import time
from watchdog.events import FileSystemEventHandler

class MyFileSystemEventHander(FileSystemEventHandler):

    def __init__(self, fn, config: Configuration):
        super(MyFileSystemEventHander, self).__init__()
        self.restart = fn
        self.config = config
        self.last = time.time()

    def on_any_event(self, event):
        # for debounce 
        cur = time.time()
        if cur - self.last < 0.25:
            return
        self.last = cur
        ext_able = False
        src = Path(event.src_path)
        if src.name  not in self.config["exclude"]:
            for ext in self.config["mon_ext"]:
                if src.suffix == ext:
                    ext_able = True
                    break

        if ext_able:
            logger.info('File changed: {}'.format(src))
            self.restart(name=src.name)


from watchdog.observers import Observer
import subprocess

class NewProcess(object):

    def __init__(self, config: dict):
        self.process = None
        self.config = config
        # self.command = self.config["cmd_args"][:]
        # self.command.insert(0, self.config["cmd"])
        # self.command[0:0] = self.config["cmd"]
        command = " ".join(self.config["cmd"])
        if len(self.config["cmd_args"]) > 0:
            self.args = "{} {}".format(command, ' '.join(self.config["cmd_args"]))
        else:
            self.args = command
        # self.args = ' '.join(self.command)

    def start(self, name=""):
        args = self.args.replace("{{name}}", name)
        logger.info('[Start process] {} ...'.format(args))
        args = args.split(" ")
        self.process = subprocess.Popen(
            args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)

    def stop(self):
        if self.process:
            logger.info('[Kill process] [{}]...'.format(self.process.pid))

            self.process.kill()
            self.process.wait()
            logger.info('[Process ended] code {}.'.format(
                self.process.returncode))
            self.process = None

    def restart(self, name=None):
        self.stop()
        self.start(name=name)

    def start_watch(self):
        observer = Observer()
        observer.schedule(MyFileSystemEventHander(self.restart, self.config), 
            path=self.config["mon_dir"], 
            recursive=self.config["recursive"]
        )
        observer.start()
        logger.info('Watching directory {}...'.format(self.config["mon_dir"]))
        if(self.config["start"]):
            self.start()
        try:
            while True:
                time.sleep(0.5)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()


np = None
def main(argv = None):
    global np
    conf = Configuration()
    if argv is None:
        argv = sys.argv[1:]
    conf.parseArgs(argv)
    
    np = NewProcess(conf.config)
    try:
        np.start_watch()
    except KeyboardInterrupt:
        print("Keyboard Interrupt, process killed")
        np.stop()
    finally:
        np.stop()


import sys
# 退出该程序时，关闭子进程
def beforeExit(signum, frame):
    print("Exit")
    global np
    np.stop()
    sys.exit()

if __name__ == '__main__':
    # import signal
    # signal.signal(signal.SIGINT, beforeExit)
    main()
    


import os
import os.path as osp

CURRENT_PATH = osp.abspath(osp.dirname(__file__))
import re
from bffacilities import createLogger
logger = createLogger("mulcpack", savefile=False, stream=True)

pat = r"-(?P<dstType>\w+)?_Qt_(?P<qt_ver>(?:\d+_){2}\d+)_(?P<compiler>[\w\d]+)_(?P<bit>\d+)bit-(?P<buildType>\w+)"

ContainedLibraries = [
    "BasicLibrary", 
    "Builib", 
    "DisplayCompass",
    "BUpdator",
]

BuildTypeMapping = {
    "Debug": "Debug",
    "MinSizeRel": "Release",
    "RelWithDebInfo": "Debug",
    "Release": "Release",
}
def extractMajorMinor(ver):
    vers = ver.split("_")
    return "_".join([vers[0], vers[1]])

class BuildTypeEntity():

    def __init__(self):
        self.qtver = None
        self.compiler = None
        self.types = []
        self._folders = []
        self.root = ""
        self.bit = 32

    def addType(self, t, file):
        if t not in self.types:
            self.types.append(t)
        self._folders.append(file)
    @property 
    def files(self):
        return self._folders

    def equal(self, qtver, compiler, bit):
        if self.bit != bit:
            return False
        if self.qtver != qtver :
            return False
        if self.compiler != compiler:
            return False
        return True

    def abspaths(self):
        p = []
        for f in self._folders:
            path = osp.join(self.root, f)
            path = path.replace("\\", "/")
            p.append(path)
        return p

    def __repr__(self):
        return f"<BTE: {self.qtver} {self.compiler} {self._folders}>"

def constructBuildType(path, **kwargs):
    bits = kwargs["bits"]
    QtCreatorPat = re.compile(f"build-{kwargs['project']}" + pat)
    target = kwargs["target"]

    buildTypes = []

    for file in os.listdir(path):
        absPath = osp.join(path, file)
        if not osp.isdir(absPath):
            continue
        matched = QtCreatorPat.match(file)
        if matched is None:
            continue
        qtver = matched.group("qt_ver")
        dstType = matched.group("dstType")
        qtver = extractMajorMinor(qtver)
        compiler = matched.group("compiler").lower()
        compBit = matched.group("bit")
        if int(compBit) not in bits:
            logger.warning(f"Path: {file} is not bits compat {compBit}, needed {bits}")
            continue
        if dstType != target:
            logger.warning(f"Path: {file} is not target type compat: {dstType}, needed {target}")
            continue

        buildType = matched.group("buildType")
        if buildType in ["MinSizeRel", "RelWithDebInfo"]:
            logger.warning(f"Found unneed type: {buildType} ")
            continue

        found = False
        for d in buildTypes:
            if d.equal(qtver, compiler, compBit):
                d.addType(buildType, file)
                found = True
        if found: continue

        btDao = BuildTypeEntity()
        btDao.bit = compBit
        btDao.root = path
        btDao.qtver = qtver
        btDao.compiler = compiler
        btDao.addType(buildType, file)

        buildTypes.append(btDao)
    return buildTypes


# exit()

from jinja2 import Template

cmakeTemp = Template(
"""
{%- for file in abspaths %}
include( "{{ file }}/CPackConfig.cmake" )
{%- endfor %}

set(CPACK_INSTALL_CMAKE_PROJECTS
    {%- for file in abspaths %}
        {%- for comp in libraries %}
    "{{ file }}/;{{ comp }};ALL;/;"
        {%- endfor %}
    {%- endfor %}
)
""")

import subprocess as sp

from argparse import ArgumentParser
def main(args):
    if len(args) == 0:
        logger.critical(f"Len args: {len(args)}, args not contain program path ")
        return
    curPath = osp.abspath(args[0])
    parser = ArgumentParser(prog="generateMultiCpack")
    parser.add_argument("--cpack", type=str, default="cpack", help="Specified CPack executable")
    parser.add_argument("--bits", type=str, default="32,64", help="Specified Bits that needs to be packed, default is 32,64")
    parser.add_argument("--target", type=str, default="Desktop", help="Specified Target system version, default: Desktop")
    parser.add_argument("--project", type=str, default="bfqtLibrary", help="Specified project name, default: bfqtLibrary")
    args = vars(parser.parse_args(args[1:]))
    
    logger.info(f"Using cpack: {args['cpack']}, destinate bits: {args['bits']}")
    ParentPath = osp.abspath(osp.join(curPath, ".."))
    
    bits = list(map(int, args["bits"].split(",")))
    args["bits"] = bits

    buildTypes = constructBuildType(ParentPath, **args)
    logger.info(f"Build Directory Found: {buildTypes}")

    libraries = ";".join(ContainedLibraries)
    projectName = args["project"]
    for bt in buildTypes:
        dirname = f"build-{projectName}-{bt.qtver}-{bt.compiler}"
        os.makedirs(dirname, exist_ok=True)
        output = cmakeTemp.render(abspaths = bt.abspaths(), folders = bt.files, libraries = ContainedLibraries)
        packName = f"MultiCPackConfig-{bt.bit}.cmake"
        with open(osp.join(curPath, dirname, packName), "w") as f:
            f.write(output)
        cwd = osp.abspath(osp.join(curPath, dirname))
        logger.info(f"CPACK DIR: {cwd}")
        p = sp.Popen([args["cpack"], "--config", packName], 
            cwd=cwd, 
            stdout=sp.PIPE, encoding='utf-8')
        print(p.stdout.read())
    # p = mp.Process("")
### cpack --config {path}/MultiCPackConfig.cmake
if __name__ == "__main__":
    CurrentPath = osp.abspath(osp.dirname(__file__))
    main([CurrentPath])


import numpy as np
from PIL import Image
from tqdm import tqdm
import datetime
import pickle
import time
import cv2 as cv
class ImagePreprocessor(object):
    """Preprocess images that contains many other areas that are not our interests"""
    def __init__(self, debug=False):
        self.debug = debug

    def setImage(self, img, dst=None):
        """ Convert other format into numpy.ndarray

        Arguments:
            img(str| np.ndarray | PIL.Image.Image | torch.Tensor)
            dst(str): directory that will be used to store processed images
        """
        self.dst = dst
        imtype = type(img)
        if imtype == str:
            self._img = np.asarray(Image.open(img))
        elif imtype == np.ndarray:
            self._img = img
        elif isinstance(img, Image.Image):
            self._img = np.asarray(img)
        elif isinstance(img, torch.Tensor):
            if img.ndim == 3:
                img = img.permute(1, 2, 0)
            ## TODO check img max val ?
            self._img = img.mul(255).byte().cpu().numpy()
        else:
            raise ValueError(f"Type Not Supported: {imtype}")
        
        assert self._img.dtype == np.uint8

        if self.debug:
            print(self._img.shape, self._img.dtype)

    def rotate(self, angle, center=None, scale=1.0, boarderValue=(0,0,0)):
        """
        Note: some problems may not be solved currently
        """
        img = self._img
        h, w = img.shape[:2]
        if center is None:
            center = (w // 2, h // 2)
        M = cv.getRotationMatrix2D(center, angle, scale)
        rotated = cv.warpAffine(img, M, (w, h), borderValue=boarderValue)
        return rotated

    def getRoiByConnected(self, thresh, areaSize):
        '''
        Arguments:
            thresh (np.ndarray): preprocessed thresh image
            areaSize(int): any components whose size is less than `areaSize` 
                will be filtered, if None, no components will be filtered
        '''
        img = self._img
        num_labels, labels, stats, centers = cv.connectedComponentsWithStats(thresh, connectivity=8, ltype=cv.CV_32S)
        image = None
        if self.debug:
            image = cv.cvtColor(img, cv.COLOR_RGB2GRAY)
            # image = np.asarray(image, dtype=np.float32)
            print(image.shape, image.ndim,image.dtype)
            # print(num_labels)
        roi_list = []
        ## TODO figure out why it starts from 1
        for t in range(1, num_labels, 1):
            x, y, w, h, area = stats[t]
            if areaSize is not None and area < areaSize:
                continue
            # save roi coordinate and width and height
            roi_list.append((x, y, w, h))
            if self.debug:
                cx, cy = centers[t]
                # 标出中心位置
                cv.circle(image, (np.int32(cx), np.int32(cy)), 2, (0, 255, 0), 2, 8, 0)
                # 画出外接矩形
                cv.rectangle(image, (x, y), (x+w, y+h), (0, 255, 0), 2, 8, 0)
                showImage(image, name="debug components", size=(w, h))
                break

        return num_labels, labels, roi_list, image

    CROP_RETAIN_WH = 0
    CROP_MIN_WH = 1
    CROP_MAX_WH = 2
    def cropByComponents(self, roi = None, \
        thresh=127, maxval=255, 
        dst_size=100000, xoff=0, yoff=0,
        crop_policy=CROP_MIN_WH, areaSize = 24000
        ):
        """
        Arguments:
            dst_size(int): final size of the image if specified
            xoff(int): offset for x
            yoff(int): offset for y

            roi (list(int)): x , y, w, h
                if roi is None: processor will try to find the first area 
                that is connected and it's size is greater than areaSize
            dst(str): image file saved location
            crop_policy(int): if 0,  width and height will be retained, 
                if greater than 0, final width and height will be the same,
                if 1, the width and height will be the smallest one
                if 2, biggest one
        """
        # im = cv.imdecode(np.fromfile(f"{src}", dtype=np.uint8), imread_type) # gray
        im = self._img
        if self._img.ndim == 3 and self._img.shape[-1] != 1:
            gray = cv.cvtColor(im, cv.COLOR_RGB2GRAY)
        else:
            gray = im

        ret, th1 = cv.threshold(gray, thresh, maxval, cv.THRESH_BINARY | cv.THRESH_OTSU )

        # contours, hierarchy = cv.findContours(im, 1, 2)
        # cnt = contours[0]
        # area = cv.contourArea(cnt)
        if roi is None:
            num_labels, labels, roi_list, image = self.getRoiByConnected(th1, areaSize=areaSize)
            x, y, w, h = roi_list[0]
        else:
            x, y, w, h = roi
        # rois.append(roi_list[0])
        if crop_policy == ImagePreprocessor.CROP_MIN_WH:
            mwh = min(w, h, dst_size)
            w = h = mwh
        elif crop_policy == ImagePreprocessor.CROP_MAX_WH:
            mwh = max(w, h, dst_size)
            w = h = mwh
        else:
            raise ValueError("Policy not supported!")
        x += xoff
        y += yoff
        cimage = im[ y:y+h, x:x+w ]

        if self.debug:
            print("ROI: ", x, y, " WH: ", w, h)
            showImage(cimage)

        self.crop_img = cimage
        self.crop_roi = [x, y, w, h] # roi_list[0]
        # return roi_list[0]

    def saveImage(self, img, dst, imsuffix=".jpg"):
        retval, buf = cv.imencode(imsuffix, cimage)
        buf.tofile(dst)

    def findEdge(self, im = None, thresh=127, maxval=255, \
            canny_thresh1=120, canny_thresh2=200, edge_weight= 0.6):
        ''' This is a test function and may change sometimes
        Arguments:
            im(np.ndarray): 
        '''
        # gray = cv.imdecode(np.fromfile(impath, dtype=np.uint8), 0)
        # gray = np.copy(self._img)
        if im is None:
            im = self._img

        if im.ndim == 3 and im.shape[-1] != 1:
            gray = cv.cvtColor(im, cv.COLOR_RGB2GRAY)
        else:
            gray = np.copy(im)
        # gray = cv.GaussianBlur(gray, (3, 3), 1)
        cimg = cv.cvtColor(gray, cv.COLOR_GRAY2BGR)
        circles = cv.HoughCircles(gray, cv.HOUGH_GRADIENT,1, 100,
                                    param1=50,param2=30,minRadius=30,maxRadius=84)
        circles = np.uint16(np.around(circles))
        for i in circles[0,:]:
            # draw the outer circle
            cv.circle(cimg,(i[0],i[1]),i[2],(0,255,0),2)
            # draw the center of the circle
            cv.circle(cimg,(i[0],i[1]),2,(0,0,255),3)

        # ret, th1 = cv.threshold(gray, thresh, maxval, cv.THRESH_BINARY | cv.THRESH_OTSU )
        th1 = cv.adaptiveThreshold(gray, maxval, cv.ADAPTIVE_THRESH_GAUSSIAN_C, cv.THRESH_BINARY, 15, 0 )
        # ret, th1 = cv.threshold(th1, thresh, maxval, cv.THRESH_BINARY | cv.THRESH_OTSU )

        # 开闭运算去除噪点
        kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, (3,3))
        th1 = cv.morphologyEx(th1, cv.MORPH_OPEN, kernel)
        th1 = cv.morphologyEx(th1, cv.MORPH_CLOSE, kernel)

        edge = cv.Canny(th1, canny_thresh1, canny_thresh2)
 
        # if self.debug:
        #     dst = cv.addWeighted(gray, (1-edge_weight), edge, edge_weight, 0)
        #     plt.figure(figsize=(10, 8 ), dpi=100)
        #     plt.imshow(dst, cmap='gray')
        #     plt.show()

        return edge, th1, cimg

    # show the colorful components
    def colorImg(self, thresh, num_labels, labels):
        '''
        thresh: 填充后的二值图
        num_labels: 连通组件的个数
        labels: 连通组件的输出标记图像，背景index=0
        '''
        # make the colors
        colors = []
        for i in range(num_labels):
            b = np.random.randint(0, 256)
            g = np.random.randint(0, 256)
            r = np.random.randint(0, 256)
            colors.append((b, g, r))
        colors[0] = (0, 0, 0)

        # draw the image
        h, w = thresh.shape
        image = np.zeros((h, w, 3), dtype=np.uint8)
        for row in range(h):
            for col in range(w):
                image[row, col] = (colors[labels[row, col]])

        return image

class CvImageReader(dict):
    """
    """
    def __init__(self, root, rootId, img_type=0, img_suffix=".png", **kwargs):
        super().__init__()
        self.root = root
        self.rootId = rootId
        self.img_suffix = img_suffix
        self.imgs = list(sorted([img for img in sorted(os.listdir(osp.join(root, rootId))) if img.endswith(self.img_suffix)]))
        self.img_type = img_type
        print("Dataset: ", len(self.imgs))

    def plt_hist(self, idxes, prange=None, density=False, stacked=False):
        """
        Arguments:
            idxes(list(int) | int)
            prange(list): plt range (two elements)
        """
        if prange is None:
            prange = [0, 256]

        if type(idxes) == list:
            imgs = []
            paths = []
            for idx in tqdm(idxes):
                img, path = self[idx]
                # img.append(img.ravel())
                # p = path[path.rfind("\\")+1:]
                # paths.append(p)
                plt.hist(img.ravel(), 256, prange, stacked=stacked, density=density)
            # print(paths, len(imgs))
            # imgs = np.asarray(imgs)
            plt.show()
        elif type(idxes) == int:
            img, path = self[idxes]
            plt.hist(img.ravel(), 256, prange, density=density)
            if not stacked:
                plt.xlabel(path)
        else:
            print("Not Supported!")

    def __getitem__(self, idx):
        if self.rootId is not None:
            path = osp.join(self.root, self.rootId, self.imgs[idx])
        else:
            path = osp.join(self.root, self.imgs[idx])
        return cv.imread(path, self.img_type), path

    def __len__(self):
        return len(self.imgs)

def saveROIOnly(src, dst = None, imread_type = 0):
    """ imread_type: IMREAD_GRAY 0 gray , IMREAD_COLOR
    """
    # src = config.DataDir / "train_raw"
    im = cv.imdecode(np.fromfile(f"{src}", dtype=np.uint8), imread_type) 
    # ret, th1 = cv.threshold(im, 127, 255, cv.THRESH_BINARY)
    th2 = cv.adaptiveThreshold(im,255,cv.ADAPTIVE_THRESH_GAUSSIAN_C,\
            cv.THRESH_BINARY, 25, 2)
    # kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, (7,7))
    # morphed = cv.morphologyEx(th2, cv.MORPH_CLOSE, kernel)
    # plt.imshow(morphed, 'gray')
    cnts = cv.findContours(th2, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)[-2]
    cnt = sorted(cnts, key=cv.contourArea)[-1]
    ## (4) Crop and save it
    x,y,w,h = cv.boundingRect(cnt)
    print(x, y, w, h)
    plt.imshow(im[y:y+h, x:x+w], 'gray')
    plt.xticks([]), plt.yticks([])
    
    plt.show()

def resizeKeepRatio(img, size, ori_size = None):
    """ori_size: None, size should be tuple
        tuple, width of size will be used (height will be ignored), keep ratio
        @retrun resized image
    """
    if ori_size is not None:
        h = ori_size[1] / ori_size[0] * size[0]
        size = (size[0], int(h))
    return cv.resize(img, size)

def showImage(img, name = "default", size = (1024, 1024), ori_size = None, skip_wait=False):
    if size is not None:
        img = resizeKeepRatio(img, size, ori_size)
    cv.imshow(name, img)
    if not skip_wait:
        cv.waitKey(0)



import unittest

class TestTemplate(unittest.TestCase):

    def __init__(self, methodName):
        super().__init__(methodName=methodName)

    def setUp(self):

        return super().setUp()

    def tearDown(self):

        return self.tearDown()

    def test_datagen(self):
        pass

if __name__ == "__main__":
    unittest.main()


'''
 # @ Author: brifuture
 # @ Create Time: 2021-04-15 12:16
 # @ Modified by: brifuture
 # @ Modified time: 2021-06-11 12:01
 # @ Description: This File is created and the program is coded by BriFuture. (c) All rights reserved.
    Provides TCP/UDP socket Server and client
 '''

import json
import socket
import select
from .utils import createLogger
logger = createLogger('sockserver', savefile=False, stream=True)
import traceback

__version__ = "0.1.7"

class EventHandler:
    def fileno(self):
        """Return the associated file descriptor"""
        raise ValueError('Not Implemented must implement')

    def handle_receive(self):
        'Perform the receive operation'
        pass

    def handle_send(self):
        'Send outgoing data'
        pass

    def handle_except(self):
        pass

    def is_valid(self):
        return True

class SocketClient(EventHandler):
    """
    1. 通过继承该类进行客户端编写，
    2. 通过 parser 回调函数进行消息读取

    Args:
        sock: socket
        address: client socket address recved from server,
            useful when other want to communicate or 
            identify the client

        except_callback: function(client: SocketClient)
    """
    def __init__(self, sock, address, **kwargs):
        self._sock = sock
        self._sock.setblocking(False)
        if address is None:
            try:
                self.address = self._sock.getsockname()
            except: 
                self.address = None
        else:
            self.address = address

        self._isValid = True

        self._parse = None
        self._server = None
        self.except_callback = kwargs.get("except_callback", None)
        self.bufferSize = kwargs.get("bufferSize", 2048)

    @staticmethod
    def createTcpClient(serverAddr):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sc = SocketClient(sock, None)
        sc._sock.setblocking(True)
        sc._server = serverAddr
        return sc

    @property
    def sock(self):
        return self._sock

    def connect(self):
        if self._server is None:
            return
        try:
            self._sock.connect(self._server)
            if self.address is None:
                self.address = self._sock.getsockname()
        except:
            logger.warning(f"[SC] Error connecting {self._server} [{self.address}]")
            self.handle_except()
        finally:
            pass

    def shutdown(self):
        self._sock.close()

    def __del__(self):
        self._sock.close()

    @property
    def parser(self):
        return self._parse

    @parser.setter
    def parser(self, parse):
        """传入一个解析器即可
        @deprecated use setParser
        ```py
        # data is a dict
        def parse(data, socketClient): print(data, socketClient)

        client.parser = parse
        ```
        """
        try:
            parse('""', self) # pass empty string
            self._parse = parse
        except:
            logger.critical(f"[TC] ***** Error testing parser, it will be ignored ****")

    def setParser(self, parser, testJson=True):
        if testJson:
            try:
                parser('""', self) # pass empty string
                self._parse = parser
            except:
                logger.critical(f"[TC] ***** Error testing parser, it will be ignored ****")
        else:
            self._parse = parser

    def is_valid(self):
        return self._isValid

    def fileno(self):
        return self._sock.fileno()

    def __repr__(self):
        return f'<Socket Client, {self._sock.fileno()}, {self.address}>'

    def handle_receive(self):
        """Handle receive data, default processor
        subclass could replace this methods to receive binary data
        """
        msg = None
        try:
            msg = self._sock.recv(self.bufferSize)
            if not msg:
                self.handle_except()
                return
        except socket.error:
            self.handle_except()
        except Exception as e:
            logger.warning(f"[TC] recv {e}")
        
        if msg is None: return

        if self._parse:
            try:
                self._parse(msg.decode(), self)
            except Exception as e:
                logger.warning(f"[TC] parse {msg}, {e}")
                traceback.print_exc()
        else:
            logger.debug(msg)

    def handle_send(self):
        pass
    def handle_except(self):
        """注意不要在 callback 中重新创建 client，否则函数调用堆栈将会溢出
        """
        self._sock.close()
        self._isValid = False
        if self.except_callback:
            self.except_callback(self)

import sys

class BaseServer():
    """ One can easily extend the base server for Tcp/Udp connection
    """
    def __init__(self):
        self._recvList = []
        self._sendList = []
        self._exceptList = []
        self.running = False

    def stop(self):
        self.running = False

    def serve_forever(self):
        self.running = True
        while self.running:
            self.serve_once(0.1)
    
    def serve_once(self, timeout=0.001):
        if len(self._recvList) == 0:
            return
        can_recv, can_send, exc = select.select(self._recvList, self._sendList, self._exceptList, timeout)
        for h in can_recv:
            h.handle_receive()
        for h in can_send:
            h.handle_send()
        for h in exc:
            h.handle_except()
        # self.checkClients() # upper level call it

class TcpSocketServer(EventHandler, BaseServer):
    """
    Args:
        ip address
        port
    """
    ClientClass = SocketClient
    def __init__(self, ip, port, maxconn = 500):
        super().__init__()
        self._server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._clients = {}
        # key: address tuple, value SocketClient
        self._recvList.append(self)
        # self._sendList = []
        self._exceptList.append(self)
        self._clientAdded = None
        self._clientRemoved = None
        try:
            self._server.bind((ip, port))
            logger.info(f"[TSS] Server Listening: {ip}:{port}")
        except Exception as e:
            logger.critical(f"[TS] Unable to listening {ip}:{port}")
            sys.exit(1)

        self._server.listen(maxconn)
        self._server.setblocking(False)
        self.maxconn = maxconn

    def setClientAddedCallback(self, p):
        """ callback """
        self._clientAdded = p

    def setClientRemovedCallback(self, p):
        """ callback """
        self._clientRemoved = p

    def fileno(self):
        return self._server.fileno()

    def needs_receive(self):
        return True

    def serve_forever(self):
        self.running = True
        count = 0
        while self.running:
            self.serve_once()
            count += 1
            if count == 6000:
                self.checkClients()

    def shutdown(self):
        self.removeAll()
        self._server.close()

    def __del__(self):
        self.removeAll()
        self._server.close()

    def checkClients(self):
        clients = list(self._clients.values())
        for c in clients:
            if not c.is_valid():
                self.removeClient(c)

    def removeAll(self):
        clients = list(self._clients.values())
        for c in clients:
            c.shutdown()
            self.removeClient(c)

    def removeClient(self, client):
        """
        Args
            client: SocketClient
        """
        if client == self: return

        logger.info(f"[TS] Removeing {client.address}")
        if client in self._recvList:
            try:
                self._recvList.remove(client)
            except: pass
        if client in self._exceptList:
            try:
                self._exceptList.remove(client)
            except: pass
        if client in self._sendList:
            try:
                self._sendList.remove(client)
            except: pass
        if client.address in self._clients:
            try:
                self._clients.pop(client.address, None)
            except: pass
            
        if self._clientRemoved:
            try:
                self._clientRemoved(client)
            except Exception as e:
                logger.warning(f"[TS] remove callback failed {e}")
    
    def _addClient(self, sock, address):
        """添加一个 客户端 socket，不可手动添加
        """

        c = TcpSocketServer.ClientClass(sock, address, except_callback=self.removeClient)
        self._clients[address] = (c)
        self._recvList.append(c)
        # self._sendList.append(c)
        self._exceptList.append(c)
        if self._clientAdded:
            self._clientAdded(c)
        else:
            logger.debug(f"[TS] client added {address}")

    def handle_receive(self):
        """接收到TCP连接
        """
        (sock, address) = self._server.accept()
        if len(self._clients) >= self.maxconn:
            # 超出最大连接数
            sock.close()
            return
        if address not in self._clients:
            self._addClient(sock, address)
        else:
            self._clients[address].sock = sock
            self._clients[address].address = address
            self._clients[address]._isValid = True
        logger.info(f'[TS] Accept From {address}, Clients Count: {len(self._clients)}')

from datetime import datetime
import threading

class UdpSocketClient(EventHandler):
    def __init__(self, port, ip = "127.0.0.1", name="default", **kwargs):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.address = (ip, port)
        self.sock.settimeout(0)
        self.sock.setblocking(False)
        self.sock.bind(self.address)
        self.bufferSize = kwargs.get("bufferSize", 1024)
        self.name = name
        self._isValid = True
        self._parse = None
        self._peer = None
        self._peers = []

        self.except_callback = kwargs.get("except_callback", None)
        # self.keepFirstPeer = False # if true, the first peer will always be used

    @property
    def parser(self):
        return self._parse

    @parser.setter
    def parser(self, parse):
        """传入一个解析器即可
        ```
        def parse(data, recver): 
            print(data, "From", recver)

        client.parser = parse
        ```
        """
        self._parse = parse

    def send(self, msg):
        """
        Args
            msg bytes
        """
        for p in self._peers:
            self.sock.sendto(msg, p)

    def is_valid(self):
        return self._isValid

    def fileno(self):
        return self.sock.fileno()

    def __repr__(self):
        return f"<UdpSocketClient {self.name} {self.address} {self.bufferSize} {self.sock.fileno()}>"

    def handle_receive(self):
        """Handle receive data, default processor
        subclass could replace this methods to receive binary data
        """
        bytesAddressPair = None
        try:
            bytesAddressPair = self.sock.recvfrom(self.bufferSize)
        except Exception as e: 
            # logger.warning(f"[UC] {self}  recv error : {e}")
            # traceback.print_exc()
            # ICMP recved cause the error, ignore it
            pass

        # message = bytesAddressPair[0]
        if not bytesAddressPair:
            # self.handle_except()
            return
        if not bytesAddressPair[1] in self._peers:
            logger.info(f"[UC] new client added: {bytesAddressPair[1]}")
            self._peers.append(bytesAddressPair[1])
        if self._parse:
            try: 
                self._parse(bytesAddressPair, self.name)
            except Exception as e:
                logger.warning(f"[UC] parse error: {e}")
                traceback.print_exc()

    def handle_send(self):
        pass

    def handle_except(self):
        self.sock.close()
        self._isValid = False
        if self.except_callback:
            self.except_callback(self)


class UdpSocketServer(BaseServer):
    """支持绑定到多个UDP端口上，使用 processor 进行回调

    Args:
        port udpsocket will bind into

    Callback:
        processor(msg: bytes, addr: tuple)

    Usage:

    ```py
    server = UdpSocketServer(bufferSize = 4096)
    
    server.createListener(10000, "server1")
    server.addProcessor(lambda x, y: print(x, y), name="server1")

    # send 
    server.sendto("server1", b"this is a test")
    server.broadcast(b"this is a broadcast")

    # in default loop
    server.serve_forever()

    # in custom loop
    while True:
        ## .. do other things
        server.serve_once()
        ## .. do other things
    ```

    """
    ClientClass = UdpSocketClient
    def __init__(self, **kwargs):
        super().__init__()
        self._listeners = {}
        self.bufferSize = kwargs.get("bufferSize", 1024)
        self._processors = {}

    @property
    def listeners(self):
        return self._listeners

    def addProcessor(self, processor, name="default"):
        if name not in self._processors:
            self._processors[name] = processor
            # logger.debug(f"[USS] add processor for {name}")
        else:
            logger.warning("[USS] dumplicated add processor.")

    def createListener(self, port, name="default", ip="127.0.0.1", processor = None):
        if name in self._listeners.keys():
            return

        assert port > 0 and port < 65535, "In valid Port"

        logger.info(f"[USS] Server Listening for '{name}': {ip}:{port}")
        listener = self.ClientClass(port, ip, bufferSize=self.bufferSize, name=name)
        listener.parser = self.recvfrom
        listener.except_callback = self.handle_except

        self._listeners[name] = listener
        self._recvList.append(listener)
        self._exceptList.append(listener)
        if processor is not None:
            self.addProcessor(processor, name)
        # self._sendList.append(listener)
        return listener
    
    def handle_except(self, client):
        logger.warning(f"[USS] Listener quit, {client.name}")
        self._listeners.pop(client.name)
        self._recvList.remove(client)
        self._exceptList.remove(client)

    def send(self, msg: bytes, name = "default"):
        """send msg to server by
        """
        if name not in self._listeners:
            return
        self._listeners[name].send(msg)

    def sendto(self, msg: bytes, peer, name="default"):
        if name not in self._listeners:
            return
        self._listeners[name].sock.sendto(msg, peer)

    def broadcast(self, msg):
        for c in self._listeners.values():
            c.send(msg)

    def recvfrom(self, byteAddrPair, name):
        """recv data from one udp server
        """
        if name in self._processors:
            # process will executed
            process = self._processors[name]
            message = byteAddrPair[0]
            address = byteAddrPair[1]
            process(message, address)
        else:
            logger.debug(f"No processor for{name}, {byteAddrPair}")



try:
    import torch
    from .torchutils import MaskContourSaver, PredictionViewer, PlotHelper
    from .datasets import CocoDataset, MaskDataset
    from .losses import HeatmapLoss
    from .models import HourGlassModule, ResidualModule, HourGlassWrapper
    from .heatmaps import HeatmapGenerator
except Exception as e :
    print("***********Error loading torch, Please check your environments**********")
try:
    from .labelmeutils import LabelmeModifier, Labelme2Vocor, Labelme2Cocoer
except Exception as e:
    print("***********Error loading labelme, Please check your environments**********")
    
from .imageutils import ImagePreprocessor



import jinja2
from jinja2 import Template
from bffacilities._constants import BFF_TEMPLATE_PATH
import os
import os.path as osp
import tqdm

tempateLoader = jinja2.FileSystemLoader(searchpath = str(BFF_TEMPLATE_PATH))
templateEnv = jinja2.Environment(loader=tempateLoader)
templateFile = "generateQtTest.jinja2"
testFileTemp = templateEnv.get_template(templateFile)
cmakeTemp = templateEnv.get_template("generateQtTestCmake.jinja2")


def writeIntoTestFile(args):
    """args
    {
        "className":  Generator,
        "file": generator,
        "methods": [ "method1", "method2" ]
    }
    """
    className = args["className"]
    testClassName = className if className.endswith("Test") else f"{className}Test"
    args["include"] = className
    args["className"] = testClassName

    if "methods" not in args:
        args["methods"] = ["method1", "method2"]
    methods = args["methods"]
    for i in range(len(methods)):
        m = methods[i]
        methods[i] = f"{m[0].upper()}{m[1:]}"
    content = testFileTemp.render(args)
    fileName = args.get("file")
    if fileName is None:
        print(content)
    else:
        filePath = osp.split(fileName)
        root = filePath[0]
        if root is None:
            root = "."
        fileName = filePath[1]
        if not fileName.startswith("tst_"):
            fileName = f"tst_{fileName.lower()}"
        if fileName.endswith("test"):
            fileName = f"{fileName}.cpp"
        if not fileName.endswith("test.cpp"):
            fileName = f"{fileName.lower()}test.cpp"
        with open(osp.join(root, fileName), "w") as f:
            f.write(content)
import clang.cindex
from clang.cindex import CursorKind, AccessSpecifier, TranslationUnit
import asciitree

class CxxParser():
    def __init__(self):
        self._methods = {}  # key is className, value is array which contained all public methods
        self.index = clang.cindex.Index.create()
        self.cur_file = None
        self.cxx_args = ["-x", "c++"]

    def parse(self, file):
        filePath = osp.split(file)
        root = filePath[0]
        self.cur_file = filePath[1]
        with open(file, "r", encoding="utf-8") as f:
            content = f.read()
            content = content.replace("signals", "private")
            content = content.replace("Q_OBJECT", "")
            content = content.replace("slots", "")
            tu = self.index.parse(self.cur_file, self.cxx_args,
                [(self.cur_file, content)])
        assert isinstance(tu, TranslationUnit)
        self.iterateNode(tu.cursor)
        # print(asciitree.draw_tree(tu.cursor, self.node_children, self.process_node))

    def _iterate(self, parent, method):
        for c in parent.get_children() :
            if c.location.file.name != self.cur_file:
                continue
            method(c)

    def iterateNode(self, node):
        p = self.processNode(node)
        if p:
            self._iterate(node, self.iterateNode)
    
    def processNode(self, node):
        kind = node.kind
        # text = node.spelling if len(node.spelling) > 0 else node.displayname
        if kind == CursorKind.CLASS_DECL:
            cn = node.spelling
            # print("\t class name ********* ", cn)
            if cn not in self._methods:
                self._methods[cn] = set()
            self._lastMethod = cn
            self._methodPublic = False
            self._iterate(node, self.processClass)
            return False
        else:
            # print(f'{kind} ==> {text} ')
            return True

    def processClass(self, node):
        kind = node.kind
        if kind == CursorKind.CXX_ACCESS_SPEC_DECL:
            declName = node.access_specifier.name
            if declName == "PUBLIC":
                self._methodPublic = True
                # print(f"\t{declName}")
            else:
                self._methodPublic = False

        elif kind == CursorKind.CXX_METHOD:
            if self._methodPublic:
                self._methods[self._lastMethod].add(node.spelling)
                # print(f' public method ==> {node.spelling} ')
            else:
                pass
                # print(f' private method ==> {node.spelling} ')
        else:
            # print(f' class {kind} ==> {node.spelling} ')
            pass
    @property
    def methods(self):
        remove = []
        for cn, ms in self._methods.items():
            if len(ms) == 0:
                remove.append(cn)
        return { k: v for k, v in self._methods.items() if k not in remove }
    def node_children(self, node):
        return list(c for c in node.get_children() 
            if c.location.file.name == self.cur_file
            )
    def process_node(self, node):
        text = node.spelling if len(node.spelling) > 0 else node.displayname
        kind = node.kind
        if kind == CursorKind.CLASS_DECL:
            if kind.name not in self._methods:
                self._methods[kind.name] = []
        if kind == CursorKind.CXX_ACCESS_SPEC_DECL:
            return f"{node.access_specifier.name}"
        return f'{kind} ==> {text} '

def saveClassTests(methods, project = None, file=None):
    # print(methods)
    print(f"Writing {len(methods)} items")

    for cn, ms in methods.items():
        args = { "className": cn, "methods": list(ms), 
            "file": cn.lower() 
            }
        if file is not None:
            args["file"] = file
        # print(args)
        writeIntoTestFile(args)
    if project is None:
        content = cmakeTemp.render(project="Unnamed", classes=methods.keys())
        print(content)
    else:
        content = cmakeTemp.render(project=project, classes=methods.keys())
        with open("Temp_CMakeLists.txt", "w") as f:
            f.write(content)
    # pass

def testClang(args):
    cparser = CxxParser()

    if args["input"]:
        file = args["input"]
        cparser.parse(file)
        # print(file, cparser.methods)
        saveClassTests(cparser.methods, project=args["project"], file=args["output"])
        return True
    elif args["dir"]:
        d = args["dir"]
        # print("============")
        exclude = args.get("exclude_dir", "")
        if exclude is not None and len(exclude) > 0:
            exclude = exclude.split(",")
        else:
            exclude = []
        exclude.append(".git")
        exclude.append("build")
        def isExcludeDir(root):
            for e in exclude:
                if e in root:
                    return True
            return False

        def isExcludeFile(file):
            if f.endswith("_p.h"):
                return True
            if f.endswith("pch.h"):
                return True
            return False
        for root, _, files in os.walk(d):
            # for d in dirs:
            #     print(root," Dir --- ", d)
            if isExcludeDir(root): 
                continue
            for f in tqdm.tqdm(files, desc=f"Searching for {root}"):
                # print(root, exclude)
                if isExcludeFile(f): continue

                if f.endswith(".h"):
                    # print("file", root, f)
                    cparser.parse(osp.join(root, f))
        saveClassTests(cparser.methods, project=args["project"])
        
        return True
    return False

def main(args = None):
    from argparse import ArgumentParser
    parser = ArgumentParser(prog="generateQtTest", description="Generate Qt Test File mannually")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-d", "--dir", type=str, help="The directory that program will walk through")
    group.add_argument("-f", "--input", type=str, help="Input content from file")
    group.add_argument("-c", "--className", type=str, help="Class Name That will be used in Test")

    parser.add_argument("-p", "--project", type=str, help="Project name if you want to write test content into CMakeLists.txt")
    parser.add_argument("-o", "--output", type=str, help="Output into file")
    parser.add_argument("-e", "--exclude-dir", type=str, help="Exclude directory, very useful when using directory `-d`")

    args = vars(parser.parse_args(args))
    if args["className"]:
        args["file"] = args["output"]
        writeIntoTestFile(args)
    else:
        if not testClang(args):
            parser.print_help()



import numpy as np
from math import pi, sqrt, atan2, asin
Rad2Deg = 180 / pi
Deg2Rad = pi / 180
# import copy
# 四元数转为旋转矩阵，将 R 系转化到 b 系的旋转矩阵
def quat2RotMat(q, mat):
    """
    q: np.ndarray(4)
    mat: np.ndarray(3)
    """
    # q: SE_q
    # # b => R
    # # row 1
    # mat[0] = 1 - 2 * q[2] * q[2] - 2 * q[3] * q[3]
    # mat[1] = 2 * (- q[0] * q[3] + q[1] * q[2])
    # mat[2] = 2 * (q[0] * q[2] + q[1] * q[3])
    # # row 2
    # mat[3] = 2 * (q[0] * q[3] + q[1] * q[2])
    # mat[4] = 1 - 2 * q[1] * q[1] - 2 * q[3] * q[3]
    # mat[5] = 2 * (- q[0] * q[1] + q[2] * q[3])
    # # row 3
    # mat[6] = 2 * (- q[0] * q[2] + q[1] * q[3])
    # mat[7] = 2 * (q[0] * q[1] + q[2] * q[3])
    # mat[8] = 1 - 2 * q[1] * q[1] - 2 * q[2] * q[2]

    # R => b
    # row 1
    mat[0] = 1 - 2 * q[2] * q[2] - 2 * q[3] * q[3]
    mat[1] = 2 * (q[0] * q[3] + q[1] * q[2])
    mat[2] = 2 * (- q[0] * q[2] + q[1] * q[3])
    # row 2
    mat[3] = 2 * (- q[0] * q[3] + q[1] * q[2])
    mat[4] = 1 - 2 * q[1] * q[1] - 2 * q[3] * q[3]
    mat[5] = 2 * (q[0] * q[1] + q[2] * q[3])
    # row 3
    mat[6] = 2 * (q[0] * q[2] + q[1] * q[3])
    mat[7] = 2 * (- q[0] * q[1] + q[2] * q[3])
    mat[8] = 1 - 2 * q[1] * q[1] - 2 * q[2] * q[2]

def eulerFromQuat(euler, q):
    pass

def eulerFromRotMat(euler, mat):
    euler[0] = atan2(mat[5], mat[8]) * Rad2Deg  # roll
    if mat[2] > 1.0:
        mat[2] = 1.0
    if mat[2] < -1.0:
        mat[2] = -1.0
    euler[1] = -asin(mat[2]) * Rad2Deg   # pitch
    euler[2] = atan2(mat[1], mat[0]) * Rad2Deg 

def rotMatFromEuler(euler):
    """
    默认使用 北东地 坐标系： 
        Heading - Z Axis
        Pitch   - Y Axis
        Roll    - X Axis
    旋转坐标系得到的旋转矩阵 （Base 坐标系到 Earch 坐标系的旋转矩阵 M）
    原向量 左乘 M.T 及则得到新坐标系下的坐标，
    在新坐标系下的向量 左乘 M 得到原坐标系下的坐标

    rot: np.ndarray (3, 3)
    
    将 Euler 角取反后相当于，旋转坐标系，最终得到的 3d 坐标是新坐标系下的坐标
    """
    y, p, r = euler * Deg2Rad
    cY = np.cos(y)
    sY = np.sin(y)
    cP = np.cos(p)
    sP = np.sin(p)
    cR = np.cos(r)
    sR = np.sin(r)
    rot = np.zeros((3, 3))
    rot[0, 0] = cY * cP
    rot[0, 1] = - sY * cR + cY * sP * sR
    rot[0, 2] = sY * sR + cY * cR * sP
    rot[1, 0] = cP * sY
    rot[1, 1] = cY * cR + sY * sP * sR
    rot[1, 2] = - cY * sR + cR * sY * sP
    rot[2, 0] = - sP
    rot[2, 1] = cP * sR
    rot[2, 2] = cP * cR
    # print(rot)
    return rot

def rotMatFromEuler2(euler):
    """
    默认使用 东北天 坐标系： 
        Heading - Z Axis
        Pitch   - Y Axis
        Roll    - X Axis
    旋转向量，坐标系不变，
    rot: np.ndarray (3, 3)
    
    将 Euler 角取反后相当于，旋转坐标系，最终得到的 3d 坐标是新坐标系下的坐标
    """
    y, p, r = euler * Deg2Rad
    cY = np.cos(y)
    sY = np.sin(y)
    cP = np.cos(p)
    sP = np.sin(p)
    cR = np.cos(r)
    sR = np.sin(r)
    rot = np.zeros((3, 3))
    rot[0, 0] = cY * cP
    rot[0, 1] = - sY * cR + cY * sP * sR
    rot[0, 2] = sY * sR + cY * cR * sP
    rot[1, 0] = cP * sY
    rot[1, 1] = cY * cR + sY * sP * sR
    rot[1, 2] = - cY * sR + cR * sY * sP
    rot[2, 0] = - sP
    rot[2, 1] = cP * sR
    rot[2, 2] = cP * cR
    # print(rot)
    return rot

def rotateVecWithEuler(vec, euler):
    """
    rotate a vector with a euler angle
    """
    rot = rotMatFromEuler(euler)
    return np.matmul(rot, vec)

if __name__ == "__main__":
    euler = np.asarray([90, 0, 45], dtype=np.float)
    res = rotateVecWithEuler((1, 0, 0), euler * -1)
    print("vec", res)


import numpy as np

class HeatmapGeneratorWithoutBatch():
    """
    Arguments: 
        out_size(int): image output resolution
        out_channels(int): model output Tensor channels, num_keypoints
    """
    def __init__(self, out_size, out_channels):
        self.out_size = out_size
        self.out_channels = out_channels
        sigma = self.out_size / 64
        self.sigma = sigma
        size = 6 * sigma + 3
        x = np.arange(0, size, 1, float)
        y = x[:, np.newaxis]
        x0, y0 = 3*sigma + 1, 3*sigma + 1
        self.g = np.exp(- ((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma ** 2))

    def __call__(self, keypoints, ratio = 1.0):
        """
        Arguments:
            keypoints: [Tensor(K, 3)] shape: N K 3
            ratio (float): generate target keypoints to dest size heatmap
        """
        batch_size = len(keypoints)
        hms = np.zeros(shape = (batch_size, self.out_size, self.out_size), dtype = np.float32)
        sigma = self.sigma
        # print("KP: ", keypoints)
        for batch_idx, kp in enumerate(keypoints): # batch kps
            # print("KP", kp)
            # assert len(kp) == 4
            for pt in kp:
                # kp torch.Tensor, size (3, )
                # print(idx, pt)
                if pt[2] > 0: 
                    x, y = int(pt[0] * ratio), int(pt[1] * ratio)
                    if x<0 or y<0 or x>=self.out_size or y>=self.out_size:
                        continue
                    topleft = int(x - 3*sigma - 1), int(y - 3*sigma - 1)
                    bottomright = int(x + 3*sigma + 2), int(y + 3*sigma + 2)

                    c,d = max(0, -topleft[0]), min(bottomright[0], self.out_size) - topleft[0]
                    a,b = max(0, -topleft[1]), min(bottomright[1], self.out_size) - topleft[1]

                    cc,dd = max(0, topleft[0]), min(bottomright[0], self.out_size)
                    aa,bb = max(0, topleft[1]), min(bottomright[1], self.out_size)
                    hms[batch_idx, aa:bb, cc:dd] = np.maximum(hms[batch_idx, aa:bb,cc:dd], self.g[a:b,c:d])
        return hms
## for debug
from .torchutils import PlotHelper
ph = PlotHelper()

class HeatmapGenerator():
    """Generate Heatmap with Batch

    Arguments: 
        out_size(int): image output resolution
        out_channels(int): model output Tensor channels
        sigma(float)
    """
    def __init__(self, out_size, out_channels, sigma = None):
        self.out_size = out_size
        self.out_channels = out_channels
        if sigma is None:
            sigma = self.out_size / 64
        self.sigma = sigma
        size = 6 * sigma + 3
        x = np.arange(0, size, 1, float)
        y = x[:, np.newaxis]
        x0, y0 = 3 * sigma + 1, 3 * sigma + 1
        self.g = np.exp(- ((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma ** 2))

    def __call__(self, keypoints, ratio = 1.0):
        """
        keypoints: target ground truth keypoints [Tensor(K, 3)] shape: N K 3
        """
        batch_size = len(keypoints)
        hms = np.zeros(shape = (batch_size, self.out_channels, self.out_size, self.out_size), dtype = np.float32)
        sigma = self.sigma
        # print("KP: ", keypoints)
        for batch_idx, kp in enumerate(keypoints): # batch kps
            # print(kp)
            for idx, pt in enumerate(kp):
                # pt.shape (3,) it is [x, y, v]
                if pt[2] == 0: continue
                x, y = int(pt[0] * ratio), int(pt[1] * ratio)
                # x, y = pt[0], pt[1]
                # if x<0 or y<0 or x>=self.out_size or y>=self.out_size:
                    # continue

                topleft = int(x - 3*sigma - 1), int(y - 3*sigma - 1) 
                bottomright = int(x + 3*sigma + 2), int(y + 3*sigma + 2)

                c, d = max(0, -topleft[0]), min(bottomright[0], self.out_size) - topleft[0]
                a, b = max(0, -topleft[1]), min(bottomright[1], self.out_size) - topleft[1]

                cc, dd = max(0, topleft[0]), min(bottomright[0], self.out_size)
                aa, bb = max(0, topleft[1]), min(bottomright[1], self.out_size)
                hms[batch_idx, idx, aa:bb,cc:dd] = np.maximum(hms[batch_idx, idx, aa:bb,cc:dd], self.g[a:b,c:d])
                # print("HM: ", aa, bb, cc, dd, "XY: ", x, y, "Ratio: ", ratio)
                # print("A, B: ", a, b)
                # ph.plotArray(hms[batch_idx, idx, :, :])
                # print("Max: ", np.max(hms[batch_idx, idx, :, :]), hms[batch_idx, idx, :, :].sum())
        # testhms = np.zeros(shape = (self.out_size, self.out_size), dtype = np.float32)
        # for hm in hms[0]:
        #     testhms += hm
        # ph.plotArray(testhms)
        return hms

import torch
import torch.nn as nn
from torch.nn import functional as F

def match_format(dic):
    loc = dic['loc_k'][0,:,0,:]
    val = dic['val_k'][0,:,:]
    ans = np.hstack((loc, val))
    ans = np.expand_dims(ans, axis = 0) 
    ret = []
    ret.append(ans)
    return ret

class HeatmapParser:
    def __init__(self):
        from torch import nn
        self.pool = nn.MaxPool2d(3, 1, 1)

    def nms(self, det):
        maxm = self.pool(det)
        maxm = torch.eq(maxm, det).float()
        det = det * maxm
        return det

    def calc(self, det):
        with torch.no_grad():
            det = torch.autograd.Variable(torch.Tensor(det))
            # This is a better format for future version pytorch

        det = self.nms(det)
        h = det.size()[-2]
        w = det.size()[-1]
        det = det.view(det.size()[0], det.size()[1], -1)
        val_k, ind = det.topk(1, dim=2)

        x = ind % w
        y = (ind / w).long()
        ind_k = torch.stack((x, y), dim=3)
        ans = {'loc_k': ind_k, 'val_k': val_k}
        return {key:ans[key].cpu().data.numpy() for key in ans}

    def adjust(self, ans, det):
        for batch_id, people in enumerate(ans): 
            for people_id, i in enumerate(people):
                for joint_id, joint in enumerate(i):
                    if joint[2]>0:
                        y, x = joint[0:2]
                        xx, yy = int(x), int(y)
                        tmp = det[0][joint_id]
                        if tmp[xx, min(yy+1, tmp.shape[1]-1)]>tmp[xx, max(yy-1, 0)]:
                            y+=0.25
                        else:
                            y-=0.25

                        if tmp[min(xx+1, tmp.shape[0]-1), yy]>tmp[max(0, xx-1), yy]:
                            x+=0.25
                        else:
                            x-=0.25
                        ans[0][0, joint_id, 0:2] = (y+0.5, x+0.5)
        return ans

    def parse(self, det, adjust=True):
        ans = match_format(self.calc(det))
        if adjust:
            ans = self.adjust(ans, det)
        return ans

from PIL import Image
from torchvision.models.detection.roi_heads import keypoints_to_heatmap
import torchvision
## for debug
# from .torchutils import PlotHelper
# ph = PlotHelper()

def translateKP(targets):
    """
        targets (list(dict))
            dict {
                "boxes": ,
                "keypoints",
            }
    """
    keypoints = []
    for gt in targets: # batch
        tmpkp = []
        for kps in gt["keypoints"]: # N points
            for kp in kps:
            # for this model we have to split all 4 keypoints separately
                tmpkp.append(kp.unsqueeze(0))
        
        # print("Tmpkp: ", tmpkp[0].shape)
        tmpkp = torch.cat(tmpkp, dim=0)
        # print("Tmpkp: ", tmpkp.shape)
        keypoints.append(tmpkp)
    # list()
    return keypoints

def np_translateKP(targets, N, K):
    keypoints = np.zeros((N, K, 2))
    for bid, gt in enumerate(targets): # batch
        for oid, kps in enumerate(gt["keypoints"]): # N points
            for kid, kp in enumerate(kps):
            # for this model we have to split all 4 keypoints separately
                # tmpkp.append(kp.unsqueeze(0))
                keypoints[bid, oid*K + kid, 0] = kp[0]
                keypoints[bid, oid*K + kid, 1] = kp[1]
    return keypoints

def heatmaps_to_keypoints(heatmaps, rois):
    """Extract predicted keypoint locations from heatmaps. Output has shape
    (#rois, 4, #keypoints) with the 4 rows corresponding to (x, y, logit, prob)
    for each keypoint.
    """
    # This function converts a discrete image coordinate in a HEATMAP_SIZE x
    # HEATMAP_SIZE image to a continuous keypoint coordinate. We maintain
    # consistency with keypoints_to_heatmap_labels by using the conversion from
    # Heckbert 1990: c = d + 0.5, where d is a discrete coordinate and c is a
    # continuous coordinate.
    offset_x = rois[:, 0]
    offset_y = rois[:, 1]

    widths = rois[:, 2] - rois[:, 0]
    heights = rois[:, 3] - rois[:, 1]
    widths = widths.clamp(min=1)
    heights = heights.clamp(min=1)
    widths_ceil = widths.ceil()
    heights_ceil = heights.ceil()
    # print("WH: ", widths, heights, widths_ceil, heights_ceil)

    num_keypoints = heatmaps.shape[1]

    if torchvision._is_tracing():
        xy_preds, end_scores = _onnx_heatmaps_to_keypoints_loop(heatmaps, rois,
                                                                widths_ceil, heights_ceil, widths, heights,
                                                                offset_x, offset_y,
                                                                torch.scalar_tensor(num_keypoints, dtype=torch.int64))
        return xy_preds.permute(0, 2, 1), end_scores

    xy_preds = torch.zeros((len(rois), 3, num_keypoints), dtype=torch.float32, device=heatmaps.device)
    end_scores = torch.zeros((len(rois), num_keypoints), dtype=torch.float32, device=heatmaps.device)

    for i in range(len(rois)):
        roi_map_width = int(widths_ceil[i].item())
        roi_map_height = int(heights_ceil[i].item())
        width_correction = widths[i] / roi_map_width
        height_correction = heights[i] / roi_map_height
        roi_map = torch.nn.functional.interpolate(
            heatmaps[i][None], size=(roi_map_height, roi_map_width), mode='bicubic', align_corners=False)[0]
        # roi_map_probs = scores_to_probs(roi_map.copy())
        ## roi_map.shape K, H, W
        w = roi_map.shape[2] ## equals to roi_map_width
        pos = roi_map.reshape(num_keypoints, -1).argmax(dim=1)

        x_int = pos % w
        y_int = (pos - x_int) // w
        # print("ROI HM to KP", x_int, y_int, w, roi_map_height)
        # print(width_correction, height_correction)

        # assert (roi_map_probs[k, y_int, x_int] ==
        #         roi_map_probs[k, :, :].max())
        x = (x_int.float() + 0.5) * width_correction
        y = (y_int.float() + 0.5) * height_correction
        xy_preds[i, 0, :] = x + offset_x[i]
        xy_preds[i, 1, :] = y + offset_y[i]
        xy_preds[i, 2, :] = 1
        end_scores[i, :] = roi_map[torch.arange(num_keypoints), y_int, x_int]

    return xy_preds.permute(0, 2, 1), end_scores


class Predict2Heatmap():
    """
     dimenstion (int): 2: 2-D heatmapt
    """
    def __init__(self, num_keypoints, dimension=2):

        self.num_keypoints = num_keypoints
        self.nums = dimension * num_keypoints

    def generate_2d_integral_preds_tensor(self, heatmaps, x_dim, y_dim):
        # assert isinstance(heatmaps, torch.Tensor)
        heatmaps = heatmaps.reshape((heatmaps.shape[0], self.num_keypoints, y_dim, x_dim))
        accu_x = heatmaps.sum(dim=2)
        accu_y = heatmaps.sum(dim=3)
        # print(heatmaps.shape, accu_x.shape, accu_y.shape)
        # accu_x = accu_x * torch.cuda.comm.broadcast(torch.arange(x_dim) \
        #     .type(torch.cuda.FloatTensor), devices=[accu_x.device.index])[0]
        # accu_y = accu_y * torch.cuda.comm.broadcast(torch.arange(y_dim) \
        #     .type(torch.cuda.FloatTensor), devices=[accu_y.device.index])[0]
        device = accu_x.device
        # accu_x is prob now
        accu_x = accu_x * torch.arange(x_dim, dtype=torch.float, device=device)
        accu_y = accu_y * torch.arange(y_dim, dtype=torch.float, device=device)
        accu_x = accu_x.sum(dim=2, keepdim=True)
        accu_y = accu_y.sum(dim=2, keepdim=True)
        # convert discrete coordinates to continuous coordinate

        return accu_x, accu_y


    def softmax_integral_tensor(self, preds, hm_width, hm_height):
        # heatmaps to keypoint
        # global soft max , batch stacks
        preds = preds.reshape((preds.shape[0], self.num_keypoints, -1))
        preds = F.softmax(preds, 2)

        # integrate heatmap into joint location
        x, y = self.generate_2d_integral_preds_tensor(preds, hm_width, hm_height)
        # x, y should be the position
        # x = x / float(hm_width) - 0.5
        # y = y / float(hm_height) - 0.5
        preds = torch.cat((x, y), dim=2)
        # print(preds)
        # preds = preds.reshape((preds.shape[0], self.nums))
        # print(preds)
        return preds


class HeatmapLoss(nn.Module):
    """
    Arguments:
        nstack(int)
        heatmapGenerator(nn.Module)

    loss for detection heatmap
    
    It's calcuates L2 distance between prediction and groundtruth
    """
    def __init__(self, heatmapGenerator, dest_size=64, \
        size_average=True, reduce=True, imsize = 256, **kwargs):
        super().__init__()
        # self.nstack = nstack
        self.generateHm = heatmapGenerator
        self.dest_size = float(dest_size)

        self.size_average = size_average
        self.reduce = reduce

        self.predict2Heatmap = Predict2Heatmap(num_keypoints=heatmapGenerator.out_channels)
        self.imsize = imsize # 256


    def weighted_mse_loss(self, input, target, weights=None):
        out = (input - target) ** 2
        if weights is not None:
            out = out * weights
        if self.size_average:
            return out.sum() / len(input)
        else:
            return out.sum()

    def weighted_l1_loss(self, input, target, weights=None):
        out = torch.abs(input - target)
        if weights is not None:
            out = out * weights
        if self.size_average:
            return out.sum() / len(input)
        else:
            return out.sum()

    def jointLocationLoss(self, preds, gt_keypoints, L1=True):
        # gt_joints_vis = args[1]
        # num_joints = int(gt_joints_vis.shape[1] / 3)
        hm_width = preds.shape[-1]
        hm_height = preds.shape[-2]
        # ph.plotTensor(preds[:1,].detach().cpu())
        pred_jts = self.predict2Heatmap.softmax_integral_tensor(preds, hm_width, hm_height)
        # print("L1 Joint: ", pred_jts)
        # print("gt: ", gt_keypoints)
        if L1:
            return self.weighted_l1_loss(pred_jts, gt_keypoints)
        else:
            return self.weighted_mse_loss(pred_jts, gt_keypoints)

    def forward(self, hm_preds, targets, rois=None, keypoint_matched_idxs=None):
        """Assume that all input image size is equal in width and height

        Arguments:
            hm_preds 
                (list(torch.Tensor)) list[N C H W], len() == Stack
                or torch.Tensor  N C H W
            targets 
                (list(dict))
                or list(torch.Tensor) 
            # image_sizes (list(torch.Tensor))
            # keypoints(list(3d array))  Tensor.size(N, K, 3)
        """
        # print(hm_preds.device, keypoints.device)
        loss_dict = {}
        if type(hm_preds) is list:
            S = len(hm_preds)
            N, K, H, W = hm_preds[0].shape
            device = hm_preds[0].device
            ratio = self.dest_size / self.imsize
            keypoints = np_translateKP(targets, N, K) * ratio
            # keypoints = self.generateHm(keypoints, ratio) # B C H W
            keypoints = torch.as_tensor(keypoints, device=device)
            # keypoints /= H
            # print("--------------")
            for i in range(S-1, -1, -1):
                preds = hm_preds[i] # B C H W
                # ph.plotArray()
                # print(preds.shape, keypoints)
                hmloss = self.jointLocationLoss(preds, keypoints)
                # print("Test", i , hmloss, "\n", keypoints)
                loss_dict[f"hmloss_{i}"] = hmloss.mean()
        else:
            N, K, H, W = hm_preds.shape
            device = hm_preds.device
            assert isinstance(hm_preds, torch.Tensor)
            assert rois is not None
            for target, roi in zip(targets, rois):
                target = target[0, :, :2] # keypoint (x, y)
                target[:, 0] -= roi[0][0]
                target[:, 1] -= roi[0][1]
                target[:, 0] = target[:, 0]# * self.dest_size / (roi[0][2]-roi[0][0])
                target[:, 1] = target[:, 1]# * self.dest_size / (roi[0][3]-roi[0][1])
            keypoints = torch.stack(targets).to(device).squeeze(1)
            keypoints = keypoints[:, :, :2] * (roi[0][2]-roi[0][0]) / self.dest_size
            # print(hm_preds.shape, keypoints.shape)
            hmloss = self.jointLocationLoss(hm_preds, keypoints)
            loss_dict["hmloss_0"] = hmloss.mean()
            ###  heatmap method @Unused
            # keypoints = self.generateHm(keypoints, ratio) # B C H W
            # # ph.plotHeatmap(hms)
            # hms = torch.as_tensor(hms)
            # # ph.plotTensor(hms)
            # hms = hms.to(device) # B C H W
            # for i in range(S-1, -1, -1):
            #     preds = combined_hm_preds[i] # B C H W
            #     # hmloss = self._forward(preds, hms)
            #     loss_dict[f"hmloss_{i}"] = hmloss.mean()
        return loss_dict
class CrossEntropyHeatmapLoss(HeatmapLoss):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def _heatmapCompare(self, pred, gt):
        """L2
        """
        # l shape: B C H W
        # print("Loss: ", pred.shape, gt.shape) # B 4 64 64
        l = torch.sqrt((pred - gt)**2)
        # l = l.mean(dim=3).mean(dim=2).mean(dim=1)
        l = l.mean(dim=3).mean(dim=2).mean(dim=1)           
        return l ## l of dim bsize
    def forward(self, ):
        S = len(hm_preds)
        N, K, H, W = hm_preds[0].shape
        device = hm_preds[0].device

        ratio = self.dest_size / self.imsize
        loss_dict = {}
        keypoints = [target["keypoints"] for target in targets]
        heatmaps = []
        roi = torch.as_tensor([[0., 0., self.dest_size, self.dest_size]]).to(device)
        for kps in keypoints:
            heatmaps_per_image, valid_per_image = keypoints_to_heatmap(kps, roi, imsize)
            heatmaps.append(heatmaps_per_image.view(-1))

        keypoint_targets = torch.cat(heatmaps, dim=0)
        # print(keypoint_targets.shape)
        for i in range(S):
            pred = hm_preds[i] # B C H W
            # print(pred.shape, N * K, H * W)
            keypoint_logits = pred.view(N * K, H * W)
            hmloss = F.cross_entropy(keypoint_logits, keypoint_targets)
            loss_dict[f"hmloss_{i}"] = hmloss
        return loss_dict

# define label
def generate_joint_location_label(config, patch_width, patch_height, joints, joints_vis):
    joints[:, 0] = joints[:, 0] / patch_width - 0.5
    joints[:, 1] = joints[:, 1] / patch_height - 0.5
    joints[:, 2] = joints[:, 2] / patch_width

    joints = joints.reshape((-1))
    joints_vis = joints_vis.reshape((-1))
    return joints, joints_vis

# define label


# define result
def get_joint_location_result(config, patch_width, patch_height, preds):
    # TODO: This cause imbalanced GPU useage, implement cpu version
    hm_width = preds.shape[-1]
    hm_height = preds.shape[-2]
    if config.output_3d:
        hm_depth = hm_width
        num_joints = preds.shape[1] // hm_depth
    else:
        hm_depth = 1
        num_joints = preds.shape[1]

    pred_jts = softmax_integral_tensor(preds, num_joints, config.output_3d, hm_width, hm_height, hm_depth)
    coords = pred_jts.detach().cpu().numpy()
    coords = coords.astype(float)
    coords = coords.reshape((coords.shape[0], int(coords.shape[1] / 3), 3))
    # project to original image size
    coords[:, :, 0] = (coords[:, :, 0] + 0.5) * patch_width
    coords[:, :, 1] = (coords[:, :, 1] + 0.5) * patch_height
    coords[:, :, 2] = coords[:, :, 2] * patch_width
    scores = np.ones((coords.shape[0], coords.shape[1], 1), dtype=float)

    # add score to last dimension
    coords = np.concatenate((coords, scores), axis=2)

    return coords



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
## 如何查找特定设备的 IP
有几种方法在局域网中找到某个设备（设为设备 A）的 IP 地址：
1. 在设备 A 上运行一段程序，该程序每隔一段时间向局域网中发送广播包（UDP 广播包），（设备 B）上运行另一个程序监听相应的端口，当接收到特定格式的消息时认为收到正确的消息，此时在命令行中打印出来的的远程设备的 IP 地址即为需要的 IP。
2. 在设备 A 上运行一段程序，该程序监听预先约定好的端口，在设备 B 上向所有 IP 地址的该端口（广播）发送消息，远程设备回复时即可得到对应的 IP。
注意发送 UDP 的广播包时，将 IP 地址设为 "255.255.255.255" 即可广播到整个网络，设为 "192.168.0.255" 可广播到 "192.168.0.0/24" 的网络。如果设为其他的 IP 地址如 “192.168.255.255” 则程序运行时会报错。
## 代码地址：
 
https://gist.github.com/BriFuture/5789fef5db9d233d2a405c0cfd6a8462

update: 19-03-01 
1. autematically check whether SERVER_RECV_PORT is useable, and if it is not useable and this script is running under server mode,
then it won't start UdpServer, 
2. replace print function with logger, when running under local mode, logger will stream message to stdout
"""
# 功能：发送广播包或接受心跳包

__version__ = "0.0.6"

SERVER_RECV_PORT = 7000
LOCAL_RECV_PORT = 7007
import socket

from bffacilities.utils import createLogger
from bffacilities._frame import Frame
logger = createLogger("broadcast.log", stream=True)

class Broadcaster(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def initArgs(self):
        from argparse import ArgumentParser
        parser = ArgumentParser(prog='bffacility broadcast', description='broadcast message to local network')
        parser.add_argument('-L', '--local', action='store_true',help='run at local client as Reciever')
        self.parser = parser

    def parseArgs(self, argv):
        args = self.parser.parse_args(argv)
        self.args = vars(args)

    def run(self):
        if self.args['local']:
            logger.info("Running Local Mode")
            local_machine()
            server = UDPServer(("", LOCAL_RECV_PORT), UdpHandler)
            try:
                server.serve_forever() 
            except Exception as e:
                logger.warning(e)
        else:
            if isServerPortAvailable(SERVER_RECV_PORT):
                logger.info("Running Server Mode")
                run_server()

def local_machine():
    """send broadcast udp packet to see who will response
    """
    con = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    con.settimeout(3)
    con.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    con.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    
    serIp = "255.255.255.255"
    con.sendto(b"'test'", (serIp, SERVER_RECV_PORT))
    logger.info("Wait Response")
    try:
        msg, addr = con.recvfrom(1024)
        logger.info(f"Recv from {addr[0]}: {msg}")
    except Exception as e:
        logger.warning(e)
    con.close()


from socketserver import UDPServer, BaseRequestHandler
import time
import json

MsgHeaderLen = len('"camera_1":')
class UdpHandler(BaseRequestHandler):
    
    def handle(self):
        msg, sock = self.request
        resp = time.ctime()
        # msgObject = json.loads(msg[MsgHeaderLen:])
        logger.info(f"recved from {self.client_address}: {msg}")
        # sock.sendto(resp.encode("ascii"), (self.client_address[0], LOCAL_RECV_PORT))
        sock.sendto(resp.encode("ascii"), self.client_address)


class Interval(object):
    def __init__(self):
        import socket
        con = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        con.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        con.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        con.settimeout(5)
        self.con = con
    
    def send(self):
        try:
            self.con.sendto(b"HeartBeat", ("255.255.255.255", LOCAL_RECV_PORT))
            msg, addr = self.con.recvfrom(1024)
            logger.info(f"Recv from {addr[0]}: {msg}")
        except Exception as e:
            logger.warning("Heart beat sent but no response: {}".format(e))

    def serve_forever(self):
        while True:
            self.send()
            time.sleep(10)


def run_server():
    from threading import Thread
    inter = Interval()
    t = Thread(target=inter.serve_forever, daemon=True)
    t.start()
    server = UDPServer(("", SERVER_RECV_PORT), UdpHandler)
    try:
        server.serve_forever() 
    except Exception as e:
        logger.warning(e)

def isServerPortAvailable(port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(3)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 0)
    sock.sendto(b"'test port available'", ("127.0.0.1", SERVER_RECV_PORT))
    try:
        msg, addr = sock.recvfrom(1024)
        logger.info("{} is not available".format(port))
    except Exception as e:
        logger.info("{} is available".format(port))
        logger.warning(e)
        sock.close()
        return True
    
    sock.close()
    return False

def main(argv = None):
    bc = Broadcaster()
    bc.initArgs()
    import sys
    if argv is None:
        argv = sys.argv[1:]
    bc.parseArgs(argv)

    bc.run()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

__version__ = "0.0.35"
__author__ = "BriFuture"

"""
0.0.35: optimize scripts in torch folder

0.0.34: fix database error

0.0.33: move logger into _globals from utils to avoid log too much content

0.0.32: update with pyupdator

0.0.22: socket server has been improved a lot, interface has changed

0.0.17: scripts in `myscripts` folder could add more scripts
without change main.py file.
"""

from .utils import createLogger, initGetText
from ._constants import BFF_ROOT_PATH, BFF_OTHER_PATH


#Dictionary with console color codes to print text
terminal_colors = {
    'HEADER' : "\033[95m",
    'OKBLUE' : "\033[94m",
    'RED' : "\033[91m",
    'OKYELLOW' : "\033[93m",
    'GREEN' : "\033[92m",
    'LIGHTBLUE' : "\033[96m",
    'WARNING' : "\033[93m",
    'FAIL' : "\033[91m",
    'ENDC' : "\033[0m",
    'BOLD' : "\033[1m",
    'UNDERLINE' : "\033[4m" 
}


#!/usr/bin/env python3
# coding:utf-8

"""windows 平台的系统栏图标
"""

import webbrowser
import os
import ctypes
from infi.systray import SysTrayIcon
from .utils import createLogger
logger = createLogger('bff_win_tray')

import locale
lang_code, code_page = locale.getdefaultlocale()

from sys import platform
if platform != "win32":
    raise TypeError('Only available on Windows ')


prototype = ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.wintypes.HWND, 
    ctypes.wintypes.LPCWSTR, ctypes.wintypes.LPCWSTR, ctypes.wintypes.UINT)
paramflags = (1, "hwnd", 0), (1, "text", "Hi"), (1, "caption", "Hello from ctypes"), (1, "flags", 0)
MessageBox = prototype(("MessageBoxW", ctypes.windll.user32), paramflags)

class WinTray():
    def __init__(self):
        icon_path = os.path.join(os.path.dirname(__file__), "imgs", "logo.ico")
        self.systray = SysTrayIcon(icon_path, "bf-facilities", 
            self.make_menu(), on_quit = self.on_quit)
            # , left_click=self.on_show, right_click=self.on_right_click)

        reg_path = r'Software\Microsoft\Windows\CurrentVersion\Internet Settings'
        self.before_quit = None
        # self.INTERNET_SETTINGS = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_ALL_ACCESS)

    def on_right_click(self):
        self.systray.update(menu=self.make_menu())
        self.systray._show_menu()

    def make_menu(self):
        return (
            ('Open', None, self.on_show), 
            ('dialogs', None, self.dialog_yes_no)
            )

    def on_show(self, widget=None, data=None):
        self.show_control_web()

    def show_control_web(self, widget=None, data=None):
        host_port = 8098
        webbrowser.open("http://127.0.0.1:%s/" % host_port)

    def on_quit(self, widget, data=None):
        # module_init.stop_all()
        if self.before_quit is not None:
            self.before_quit()

    def serve_forever(self):
        self.systray.start()

    def dialog_yes_no(self, msg="msg", title="Title", data=None, callback=None):
        res = MessageBox(text = "msg", caption = "title", flags = 1)
        # Yes:1 No:2
        if callback:
            callback(data, res)
        return res

sys_tray = None

def main(arg):
    global sys_tray
    sys_tray = WinTray()
    ctypes.windll.user32.ShowWindow(ctypes.windll.kernel32.GetConsoleWindow(), 0)
    sys_tray.serve_forever()

if __name__ == '__main__':
    main()




import platform
import os
import atexit
import logging
from pathlib import Path

__version__ = "0.1.1"

_LOG_DIR = Path.home() / ".bfp" / ".logs"   # Log Path
_LOG_DIR.mkdir(parents=True, exist_ok=True)

formaterStr = "%(asctime)s %(levelname)s:  %(message)s"

def createLogger(name: str, savefile = True, stream = False, 
    level = logging.INFO, basedir = None,  **kwargs):
    """create logger
    Args: 
        name : suffix will be appended, for example, `test` will be `test.log`
        basedir: default is `~/.logs`

    kwargs:
        timeformat: default is "%Y-%m-%d %H:%M:%S" 
    :: logger_prefix deprecated ::

    """

    logger = logging.getLogger(name)
    # print("create logger", name)
    # import traceback
    # traceback.print_stack()
    if hasattr(logger, "_bf_inited"):
        # return created logger
        # print("Created Logger Returned")
        return logger

    logger._bf_inited = True

    _formater = logging.Formatter(formaterStr, kwargs.get("timeformat", "%Y-%m-%d %H:%M:%S"))

    logger.setLevel(level)
    if savefile:
        if basedir is None:
            basedir = _LOG_DIR
        elif type(basedir) == str:
            basedir = Path(basedir)
        os.makedirs(basedir, exist_ok=True)
        if not name.endswith(".log"):
            name = name + ".log"
        log_file = basedir / name
        fh = logging.FileHandler(log_file)
        fh.setFormatter(_formater)
        fh.setLevel(level)
        logger.addHandler(fh)
        # print("add file handler", log_file)
    if stream:
        sh = logging.StreamHandler()
        sh.setLevel(level)
        sh.setFormatter(_formater)
        logger.addHandler(sh)
    return logger

def changeLoggerLevel(logger, level):
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)

## ********** For Translating **************
_LANGUAGE_DIR = (Path(__file__).parent / "locale").resolve()
import gettext
def initGetText(domain="myfacilities", dirs = _LANGUAGE_DIR) -> gettext.gettext:
    """make some configurations on gettext module 
    for convenient internationalization
    """
    gettext.bindtextdomain(domain, dirs)
    gettext.textdomain(domain)
    gettext.find(domain, "locale", languages=["zh_CN", "en_US"])
    return gettext.gettext
## ********** For Translating **************

def lockfile(fileName, start = None, stop = None, logger=None):
    """
    original : Fix Multiple instances of scheduler problem  
        https://github.com/viniciuschiele/flask-apscheduler/issues/51
    :param str filename: specified lock filename, such as app.lock
    :param function start: callback for app loop start
    :param function stop: callback for app loop stop
    :@Return True success
    """
    if logger is None:
        logger = logging.getLogger("bffacilities")
    if platform.system() != 'Windows':
        fcntl = __import__("fcntl")
        f = open(fileName, "wb")
        try:
            fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            if start is not None:
                start()
            # logger.debug("Scheduler Started...")
        except Exception as e:
            logger.error('Exit the scheduler, error - {}'.format(e))
            if stop is not None:
                stop()
            return False
        def unlock():
            fcntl.flock(f, fcntl.LOCK_UN)
            f.close()
        atexit.register(unlock)
    else:
        msvcrt = __import__("msvcrt")
        f = open(fileName, "wb")
        logger.info("Lock file is: ", os.path.realpath(f.name))
        try:
            msvcrt.locking(f.fileno(), msvcrt.LK_NBLCK, 1)
            if start is not None:
                start()
            # logger.debug("Scheduler Started...")
        except Exception as e:
            logger.error('Exit the scheduler, error - {}'.format(e))
            if stop is not None:
                stop()
            return False
        def _unlock_file():
            try:
                f.seek(0)
                msvcrt.locking(f.fileno(), msvcrt.LK_UNLCK, 1)
                f.close()
                os.remove(f.name)
            except IOError:
                raise
        atexit.register(_unlock_file)
    return True

import hashlib
def simplePasswordEncry(password):
    t = hashlib.sha256(password.encode())
    d = t.digest()
    return ''.join(('{:02x}'.format(x) for x in d))

import subprocess
from ._constants import BFF_OTHER_PATH
def createShortCut(target, link = None, desc="", workingdir=None, desktop=None, startup=None):
    """target: Path
    """
    scriptPath = BFF_OTHER_PATH / "winbat" / "createshortcut.js"
    assert link is not None or desktop is not None or startup is not None
    if workingdir is None:
        assert isinstance(target, Path)
        workingdir = (target / "..").resolve()
    script = ["wscript.exe", str(scriptPath), 
        "--workingdir", str(workingdir),
        "--target", str(target),
        "--desc", desc,
        ]
    if desktop is not None:
        script.append("--desktop")
        script.append(desktop)
    if startup is not None:
        script.append("--startup")
        script.append(startup)

    try:
        process = subprocess.Popen(script, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # ret = p.stdout.read().decode()
    except Exception as e:
        logger = logging.getLogger("bffacilities")
        logger.warn(f"Error when creating shortcut {e} for {script}")



"""
Used as common Frame for myscripts
"""
class Frame(object):
    def __init__(self):
        self.config = {}

    def readConfig(self, file):
        pass

    def initArgs(self):
        """ @return ArgumentParser """
        pass

    def parseArgs(self, argv):
        # args = self._initArgs().parse_args()
        pass
import sys
def common_entry(frame):
    def entry_main(argv = None):
        f = frame()
        f.initArgs()
        if argv is None: argv = sys.argv[1:]
        f.parseArgs(argv)
    return entry_main


from __future__ import print_function

import argparse
import glob
import os
import os.path as osp
import sys

import imgviz
import numpy as np

import labelme

import json
import shutil as sh
from .labelmeutils import Labelme2Vocor

def main_labelme2voc(args, ordered_keys=None):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("input_dir", help="input annotated directory")
    parser.add_argument("output_dir", help="output dataset directory")
    parser.add_argument("--labels", help="labels file", required=True)
    parser.add_argument(
        "--noviz", help="no visualization", action="store_true"
    )
    args = parser.parse_args(args)

    vocor = Labelme2Vocor(args.input_dir, args.output_dir, args.labels, noviz=args.noviz, debug=False)
    vocor.getClasses()
    vocor.output(ordered_keys=ordered_keys)


from .labelmeutils import Labelme2Cocoer

def getParser_labelme2coco():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("input_dir", help="input annotated directory")
    parser.add_argument("output_dir", help="output dataset directory")
    parser.add_argument("--labels", help="labels file", required=True)
    parser.add_argument("--debug", action="store_true", default=False, help="labels file")
    return parser


def _main_labelme2coco(arguments=None):
    """@deperacted 
    copy following scripts in your data directory 
    and use setClassId of cocoGener instead
    """
    parser = getParser_labelme2coco()
    if arguments is None:
        arguments = sys.argv[1:]
    args = parser.parse_args(arguments)

    if osp.exists(args.output_dir):
        print("Output directory already exists: ", args.output_dir)
        sh.rmtree(args.output_dir)
        # sys.exit(1)
    os.makedirs(args.output_dir)
    os.makedirs(osp.join(args.output_dir, "JPEGImages"))
    print("Creating dataset:", args.output_dir)

    label_files = glob.glob(osp.join(args.input_dir, "*.json"))

    cocoGen = Labelme2Cocoer(args.output_dir, debug=args.debug)
    cocoGen.classNameToId(args.labels)
    cocoGen.generateCocoJson(label_files)
    cocoGen.output()
    return cocoGen


"""
TODO reformat this file and imageutil, labelme utils
"""


import cv2 as cv
import numpy as np
import os
import matplotlib.pyplot as plt
import os.path as osp
from .labelmeutils import encodeImage

import torch

class PlotHelper():
    """Helper used to plot serveral images into one figure
    """
    def __init__(self, vmin = 0, vmax = 255):
        self.vmin = vmin
        self.vmax = vmax
        
    def checkzeros(self, imglist, imgstart = 0):
        channels = len(imglist)
        zero = []
        for i in range(channels):
            imgploti = imglist[i]
            # print(imgploti)
            if np.count_nonzero(imgploti) == 0:
                zero.append(i + imgstart)

        if len(zero) > 0:
            print(f"Following (total: {len(zero)}) are all zeros: ", zero)

        self.zero = zero
        return len(zero)

    def _plot(self, imglist, rows=2, cols=2, dpi=60, imgstart=0, names=None, cmap='gray'):
        """
        imglist(list(np.ndarray)):
        """

        fig = plt.figure(figsize=(18, 16), dpi= dpi, facecolor='w', edgecolor='w', clear=True)
        imgstart = imgstart * (rows * cols)
        channels = len(imglist)
        # channels = imglist.shape[0]
        if imgstart + rows * cols > channels:
            print(f"Index Invalid: {imgstart} - {rows * cols}")
            # imgstart = 0

            # names = [f"{i}" for i in range(rows * cols)]
        for idx in range(imgstart, imgstart + rows * cols): #imglist.shape[0]):
            if idx >= channels: break

            a = fig.add_subplot(rows, cols, idx - imgstart +1)
            # Turn off tick labels
            a.set_yticklabels([])
            a.set_xticklabels([])
            if names is not None:
                a.set_xlabel(names[idx])
            # plt.axis('off')
            imgploti = imglist[idx]
            imgploti = plt.imshow(imgploti, cmap=cmap) # * 255, , vmin=0, vmax=255)


    def plotImgList(self, imglist, checkzeros=True, names=None, **kwargs):
        """
        Arguments:
            imglist (np.ndarray | list | torch.Tensor)
        """
        imgtype = type(imglist)
        if imgtype == np.ndarray:
            shape = imglist.shape
            ndim = imglist.ndim 
            if ndim == 2:
                imglist = [imglist]
            elif ndim == 3:
                imglist = [img for img in imglist]
            else:
                raise ValueError(f"Unsupported ndarray shapeP: {shape} {ndim}")
        elif imgtype == list:
            pass
        elif imgtype == torch.Tensor:
            assert len(imglist.shape) == 4  ## B C H W
            imglist = imglist[0,:,:,:].cpu()
            print("Plot on converted Tensor: ", imglist.shape)
            ims = [imglist[i].cpu().numpy() for i in range(len(imglist))]
        else:
            raise ValueError(f"Unsupported type: {imgtype}")
        # print(imglist)
        if names is not None:
            assert len(names) >= len(imglist)
        if checkzeros:
            self.checkzeros(imglist, kwargs.get("imgstart", 0))

        self._plot(imglist, names=names, **kwargs)
        
    def plotTensor(self, imtensor, cmap='gray', **kwargs):
        """Tensor
        """
        tlen = len(imtensor.shape)
        if tlen == 3:
        # assert imtensor.shape[0] == 3
            im = imtensor.permute(1, 2, 0).mul(256).byte().cpu().numpy()
            self.plotArray(im, cmap, **kwargs)
        elif tlen == 4:
            ## channel may not be 3
            # print(imtensor.shape)
            for imt in imtensor:
                merged = np.zeros((imt.shape[-2], imt.shape[-1]))
                for im in imt:
                    im = im.cpu().numpy()
                    self.plotArray(im, cmap, **kwargs)
                    # merged = np.where(im > 0, im, merged)
                # self.plotArray(merged, cmap, **kwargs)

    def plotArray(self, im, cmap='gray', figsize=(12, 8), **kwargs):
        # print(im.shape)
        fig = plt.figure(figsize=figsize, dpi= 100, facecolor='w', edgecolor='w', clear=True)
        plt.imshow(im, cmap=cmap)
        plt.show()

    def plotHeatmap(self, imarr, cmap='gray', **kwargs):
        # for batch in imarr:
        hms = imarr[0]
        for hm in hms:
            self.plotArray(hm)

    def drawPoints(self, points, size=None, img=None, bright=2.5, cmap='gray', **kwargs):
        """
            points (list(np.ndarray))
            size (tuple(height, width))
        """
        # assert len(size) == 3
        assert size is not None or img is not None

        if img is None:
            img = np.zeros(size, dtype=np.float)
        else:
            img = img.copy()
        # print(img.shape)
        for po in points:
            x, y = po[:2]
            x, y = int(x), int(y)
            img[x, y] = bright
        self.plotArray(img, cmap=cmap, **kwargs)
        

from PIL import Image, ImageDraw
import imgviz
import types
import json
import shutil as sh

class MaskContourSaver():
    """
    Used to save mask contour predicted by ai. 
    @Note it depends on certain model and datasets, and outpus only mask, 
        use PredictionViewer instead, which provides keypoint label output
    @See torchutils.PredictionViewer

    Arguments:
        labels (function , list) it maps the idx of predicted results into Readable texts

        root(str): root dir contains images, should be different from dst_dir
    """

    def __init__(self, labels=None, root = None, img_suffix=".jpg", debug=False, **kwargs):
        self.debug = debug
        self.root = root
        # self.imgs = sorted([file for file in os.listdir(root) if file.endswith(img_suffix)])
        self.img_suffix = img_suffix

        if type(labels) is list:
            self._captions = labels
        else:
            label = []
            assert isinstance(labels, types.FunctionType)
            try:
                for i in range(1000):
                    label.append(labels(i))
            except: pass
            self._captions = label
        if self.root is not None:
            print("Root: ", self.root)
        print("Get captions: ", self._captions)


    def _save(self, im, dst_dir, imname, shapes, imsize):
        """
        im(np.ndarray | PIL.Image.Image)
        imname(str): image name without suffix, such as `1.jpg` will get `1`
        shapes(list(dict)):
        """
        imfilename = f"{imname}{self.img_suffix}"
        ## @See Also encodeImage
        # imageData = LabelFile.load_image_file(imgloc)
        # imageData = base64.b64encode(imageData).decode("utf-8")
        imageData = encodeImage(im, self.img_suffix)

        data = {"version": "4.5.5",
            "flags": {}, "shapes":shapes, "imagePath": imfilename, 
            "imageHeight": imsize[1],
            "imageWidth": imsize[0],
            "imageData": imageData
        }
        if self.root is not None:
            try:
                imgloc = osp.join(self.root, imfilename)
                sh.copy(imgloc, osp.join(dst_dir, imfilename))
            except Exception as e:
                print(e)
        jsonloc = osp.join(dst_dir, imname+".json")
        with open(jsonloc, "w") as f:
            f.write(json.dumps(data, indent=True))
    
    def _getPrediction(self, idx):
        """
        @Returns 
            True prediction is valid
        """
        img, _ = dataset[idx]
        imsize = img.shape[-2], img.shape[-1]
        img = img.to(device)
        predictions = self.model([img])
        prediction = predictions[0]  ## batch size is one

        self.boxes = prediction["boxes"]
        self.labels = prediction["labels"].cpu().numpy()
        self.masks = prediction['masks'].mul(255).byte().cpu().numpy()
        
        self.maskcount = masks.shape[0] if maskperimage <= 0 else maskperimage
        
    def save(self, dst_dir, dataset, imgnames, drange=None,\
            maskperimage=-1, skipUnmask=True,\
            contourDis = 8, contourGap=4, contourArea=None, contourThresh=None, contourSaveMax=False, \
            device=torch.device('cpu'), **kwargs):
        """
        Before invoking this method, model should be in `eval` mode, `with torch.no_grad()` is 
        necessary.

        Arguments:
            dataset(torch.utils.Dataset): 
                dataset should return tuple (img, target)
                Generally property: 
                1. imgs, which provide the names of images, it should be passed as parameter `imgnames`
                2. root, root_dir indicates where the data is from, it should be passed to constructor
                3. Labels, which could be used to repreter label index

            imgnames(list):
            
            drange(range):

            maskperimage(int): if -1, then no mask limit for each image

            contourThresh(tuple): [threshold, maxval, imtype]
            contourArea(int): if none, contour area will not be accounted for. Otherwise any contour whose are is less than `contourArea` will be filtered
            contourGap(int): 
            contourSaveMax(bool): Not Implemented Yet
        """
        assert dst_dir != self.root
        # device = torch.device(device)
        # assert device == self.model.device
        os.makedirs(dst_dir, exist_ok=True)
        
        if drange == None:
            drange = range(0, len(dataset))
        
        assert drange[-1] <= len(dataset)

        if contourThresh is None:
            contourThresh = [127, 255, 0]
        
        assert len(contourThresh) == 3
        print("Start prediction on: ", drange, "...")
        self.contour_areas = []
        self.skipped_idxes = []
        for idx in drange:
            self._getPrediction(idx)
            if self.maskcount == 0:
                continue

            found_contours = []
            contour_areas = []
            for c in range(self.maskcount):
                maskdata = self.masks[c, 0]
                label_name = self._captions[self.labels[c]]

                ret, thresh = cv.threshold(maskdata, contourThresh[0], contourThresh[1], contourThresh[2])
                contours, hierarchy = cv.findContours(thresh, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)

                if len(contours) == 0:
                    continue
                # reshape as a list, the first contour will be used
                contour = contours[0]
                actualarea = cv.contourArea(contour)

                if contourArea is not None and actualarea < contourArea:
                    # print("ContourArea is too small: ", actualarea)
                    continue
                
                if self.debug:
                    print("Actual contour area: ", actualarea)
                contour = contour.reshape(-1, 2)
                contour = cv.convexHull(contour)
                ## disable contourGap using hull instead
                if len(contour) > contourGap * 5:
                    contour = contour[0::contourGap]
                if len(contour) < 2:
                    continue
                found_contours.append(contour)
                contour_areas.append(actualarea)

            if len(contour_areas) > 0:
                self.contour_areas.append(contour_areas)
            shapes = []
            for contour in found_contours:
                shapes.append({"label": label_name, 
                    "points": contour.tolist(),       
                    "group_id": None,
                    "shape_type": "polygon",
                    "flags": {}
                })
            if len(shapes) == 0 and skipUnmask:
                self.skipped_idxes.append(idx)
                continue
            print(f"Idx: {idx}: mask will be saved: {len(shapes)}")

            imfilename = imgnames[idx]
            imname = imfilename[:imfilename.rfind(self.img_suffix)]
            self._save(osp.join(self.root, imfilename), dst_dir, imname, shapes, imsize)
            print("==============================")
            # if self.debug:
            #     break
        
    @staticmethod
    def saveContour(prediction, dataset, idx, saveLoc, labelList = None,\
        gap=4, size=(512, 512), img_suffix=".jpg"):
        """
        Deprecated, Use Saver instead
        Save Contour By AI prediction and CV contours

        Arguments:
            dataset(torch.utils.Dataset)
            idx(int)
            saveLoc(str)
            labels(list)
        """
        os.makedirs(saveLoc, exist_ok=True)
        boxes = prediction["boxes"]
        labels = prediction["labels"]
        masks = prediction['masks']
        i = 1
        count = masks.shape[0]
        label = labels.cpu().numpy()
        shapes = []
        if labelList is None:
            labelList = dataset.Labels
        print(f"{idx} Masks: {count}")

        imname = dataset.imgs[idx]
        imfilename = imname[:imname.rfind(img_suffix)]
        for c in range(count):
            mask = masks[c, 0]
            label_name = labelList[label[c]]
            imdata = mask.mul(255).byte().cpu().numpy()
            # im = Image.fromarray(imdata)
            # display(im)
            ret, thresh = cv.threshold(imdata, 127, 255, 0)
            contours, hierarchy = cv.findContours(thresh, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)
            if len(contours) == 0:
                continue
            # reshape as a list
            contour = contours[0].reshape(-1, 2)
            if len(contour) > gap * 5:
                contour = contour[0::gap]
            if len(contour) < 2:
                continue
            shapes.append({"label": label_name, "points": contour.tolist(),       
                "group_id": None,
                "shape_type": "polygon",
                "flags": {}
            })
        imageData = LabelFile.load_image_file(osp.join(dataset.root, imname))
        imageData = base64.b64encode(imageData).decode("utf-8")
        data = {"version": "4.5.5",
            "flags": {}, "shapes":shapes, "imagePath": imname, 
            "imageHeight": size[1],
            "imageWidth": size[0],
            "imageData": imageData
        }
        with open(osp(saveLoc, imfilename+".json"), "w") as f:
            f.write(json.dumps(data, indent=True))

    @staticmethod
    def mergeLabelIntoImages(labeldir, imagedir):
        labels = os.listdir(labeldir)
        for label in labels:
            if not label.endswith(".json"):
                raise ValueError("Labels should be json file")
            path = osp.join(imagedir, label)
            if osp.exists(path):
                print("Removing: ", path)
                os.remove(path)
            sh.move(osp.join(labeldir, label), imagedir)

        sh.rmtree(labeldir)
import math
class PredictionViewer(MaskContourSaver):
    """
    1. Display model's prediction 
    2. Save prediction as labeled json files or visualized images

    Arugments:
    """

    def __init__(self, labels=None, img_suffix=".jpg", debug=False, kpcaptions=None, **kwargs):
        super().__init__(labels=labels, img_suffix=img_suffix, debug=debug, **kwargs)
        self.kpcaptions = kpcaptions
        if kpcaptions is not None:
            print("KP captions:", kpcaptions)

    @staticmethod
    def mask_to_box(mask):
        assert type(mask) is np.ndarray
        pos = np.where(mask)
        xmin = np.min(pos[1])
        xmax = np.max(pos[1])
        ymin = np.min(pos[0])
        ymax = np.max(pos[0])
        return [xmin, ymin, xmax, ymax]

    def setPrediction(self, predictions, mask3channels=True, \
        score_threshold=0.5, hint=True):
        """Note, if predictions is a tensor contains multiple predictions, the first one will be used
            and others will be droped.
            score_threshold is deprecated, if you want to filter that is not clearly identified,
            use `box_thresh` for rpn.
        """

        self.score_threshold = score_threshold

        if type(predictions) == list:
            if len(predictions) == 0:
                raise ValueError("No Prediction found")
            prediction = predictions[0]
        else:
            prediction = predictions

        self.boxes =  prediction["boxes"].cpu().numpy()
        self.labels = prediction["labels"].cpu().numpy()
        self.scores = prediction["scores"].cpu().numpy()

        # proposals = []
        # for i, score in enumerate(self.scores):
        #     # print("Score", i, score)
        #     if score > self.score_threshold:
        #         proposals.append(i)
        # self.boxes = np.take(self.boxes, proposals, axis=0)
        # print(self.boxes)

        if "masks" in prediction:
            masks = prediction['masks']
            masks = masks.permute(0, 2, 3, 1)#.squeeze(0)
            ## N H W C
            # if mask3channels:
            #     if masks.shape[1] == 1:
            #         masks = masks.repeat(1, 1, 1, 3)
            #     # now masks showld be N H W C
            # else:
            #     masks = masks.squeeze(3)
                # now masks showld be N H W
            # maskimg = []
            # for mask in masks:
                # maskim.append(mask.mul(128).byte().cpu().numpy())
                # maskimg.append(Image.fromarray(mask.mul(128).byte().cpu().numpy()))
            # self.masks = maskimg
            self.masks = masks.mul(255).byte().cpu().numpy()
            # print(self.masks.shape)
        else:
            self.masks = None
        if "keypoints" in prediction:
            keypoints = prediction["keypoints"]
            self.keypoints = keypoints.cpu().numpy()
            self.keypoints_scores = prediction["keypoints_scores"].cpu().numpy()
            # N 2
        else:
            self.keypoints = None

        # print(self.masks.shape, self.keypoints.shape)

        if hint:
            m = self.masks.shape if self.masks is not None else 0
            k = self.keypoints.shape if self.keypoints is not None else 0
            print("Found masks", m, " Keypoints: ", k)

    def convertImg(self, imarr):
        """
        Argument:
            imarr (np.ndarray) # ndim == 3 (H, W, C) channels == 3
        Return:
            np.ndarray (H, W, C)
        """
        imtype = type(imarr)
        # print(imtype)
        if imtype == np.ndarray:
            pass
        elif isinstance(imarr, torch.Tensor):
            assert len(imarr.shape) == 3
            imarr = imarr.permute(1, 2, 0)
            if imarr.shape[-1] == 1:
                imarr = imarr.repeat(1, 1 ,3)
            imarr = imarr.mul(255).byte().cpu().numpy()
        elif isinstance(imarr, (Image.Image, )):
            # im = Image.fromarray(imarr, mode="RGB")
            im = imarr
            imarr=np.asarray(imarr)
            if imarr.ndim != 3:
                imarr = np.asarray([imarr])
                imarr = np.repeat(imarr, 3, axis=0)
                imarr = imarr.swapaxes(0, 1)
                imarr = imarr.swapaxes(1, 2)
        else:
            raise ValueError("Unsupported Type", imtype)
        
        assert imarr.ndim == 3 and imarr.shape[-1] == 3
        # print("Show: ", imarr.shape)

        return imarr

    def showMask(self, imarr, box_swap_xy=True,
        # mask
        mask_threshold=30, mask_alpha=0.25, box_only=False):
        """
        @return np.ndarray
        """
        imarr = self.convertImg(imarr)

        captions = self.caption_with_score

        # print(self.boxes, captions)
        colormap = imgviz.label.label_colormap()
        if self.masks is None or box_only:
            # data = imgviz.data.arc2017()
            # bboxes = self.boxes
            if box_swap_xy:
                bboxes = []
                for box in self.boxes:
                    bboxes.append((box[1], box[0], box[3], box[2]))
            else:
                bboxes = self.boxes
            viz = imgviz.instances2rgb(imarr, bboxes=bboxes, \
                labels=self.labels, captions=captions, \
                colormap=colormap, alpha=mask_alpha, \
                font_size=12, line_width=1)
        else:
            masks = self.masks > mask_threshold
            # mask didnot need all 3 channels
            masks = masks[:, :, :, 0]
                # masks = np.squeeze(masks, axis=3)
            viz = imgviz.instances2rgb(imarr, masks=masks, \
                labels=self.labels, captions=captions, \
                colormap=colormap, alpha=mask_alpha, \
                font_size=12, line_width=1)
        return viz

    def showKeypoint(self, im, radius=5, kp_fill=None, num_color=248, **kwargs):
        """
        Arugments:
            im(PIL.Image.Image)
        """
        if self.keypoints is None:
            print("Unable to find keypoint in prediction output.")
            return None

        assert isinstance(im, Image.Image)
        colormap = imgviz.label.label_colormap()
        imdraw = ImageDraw.Draw(im)
        keypoints = self.keypoints[:,:,:2]
        i = 1
        rgbmode = im.mode == 'RGB'
        
        print(keypoints)

        for kps in keypoints:
            for k in kps:
                # print(fill, k)
                if rgbmode:
                    fill = tuple(x for x in colormap[num_color-i-1]) if kp_fill is None else kp_fill
                else:
                    fill = colormap[num_color-i-1][0] if kp_fill is None else kp_fill

                if radius > 0:
                    imdraw.rectangle([k[0] - radius, k[1] - radius, k[0]+radius, k[1]+radius], fill=fill)
                else:
                    imdraw.point([k[0], k[1]], fill=fill)
                i+=1
        return im

    @property
    def caption_with_score(self):
        return [f"{self._captions[i]}: {score:.2f}" for i, score in zip(self.labels, self.scores)] #if score > self.score_threshold

    def show(self, imarr, box_swap_xy=True,
        # mask
        mask_threshold=30, mask_alpha=0.25,
        # keypoint
        radius=5, kp_fill=(128, 200, 128), **kwargs
        ):
        """Draw masks and keypoints(if exists) on the image

        @Note setPrediciton must be called before this method invoked
        
        Arguments:
            imarr (np.ndarray) # ndim == 3 (H, W, C) channels == 3
            box_swap_xy(bool), the imgviz tool xy axis seems different from the screen axis,
                so it will be swapped if necessary.

        Return:
            im (Image.Image)
        """
        imarr = self.convertImg(imarr)
        # swap x, y
        viz = self.showMask(imarr, box_swap_xy, mask_threshold=mask_threshold, mask_alpha=mask_alpha)
        im = Image.fromarray(viz)
        im = self.showKeypoint(im, radius=radius, kp_fill=kp_fill)
        return im

    def displayWrongHint(self, predictions, idx=None, showWrongHintOnly=True, expected_box_num=2):    
        self.setPrediction(predictions, hint=False)
        box_num = len(self.boxes)
        if box_num != expected_box_num:
            print(f"{idx} Wrong predicted box: ", box_num)
        elif not showWrongHintOnly:
            print("Predicted box: ", box_num)

    Save_Success = 0
    Save_SkippedNoMask = 1
    Save_NoMaskPredicted = 2

    def _maskToContour(self, found_contours, contourGap, contourThresh, contourArea):
        if self.masks is None:
            return

        maskNum = self.masks.shape[0] # 

        for ci in range(maskNum):
            label_name = self._captions[self.labels[ci]]
            maskdata = self.masks[ci]
            ret, thresh = cv.threshold(maskdata, contourThresh[0], contourThresh[1], contourThresh[2])
            # thresh = cv.adaptiveThreshold(maskdata, 255, cv.ADAPTIVE_THRESH_MEAN_C, cv.THRESH_BINARY, 11, 2)
            # if self.debug:
            # display(Image.fromarray(thresh))
            contours, hierarchy = cv.findContours(thresh, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)
            # print("fondContours", maskdata.shape, label_name, len(contours))
            if len(contours) == 0:
                continue
            # reshape as a list, the first contour will be used
            newcontours = []
            for contour in contours:
                actualarea = cv.contourArea(contour)

                if contourArea is not None and actualarea < contourArea:
                    # print(f"[{label_name}] ContourArea is too small: ", actualarea)
                    continue

                contour = contour.reshape(-1, 2) # two-d array is needed for labelme
                # print("Contour ", len(contour))
                conlen = len(contour)
                if conlen < 2:
                    continue
                ncontour = []
                i = 0
                x1 = contour[i]
                ncontour.append(x1)
                i += 1
                while i < conlen - 1: 
                    x2 = contour[i]
                    dis = (x1[0] - x2[0]) ** 2 + (x1[1] - x2[1]) ** 2
                    # print(dis)
                    dis = math.sqrt(dis)
                    if dis >= contourGap or contourGap <= 0:
                        x1 = x2
                        ncontour.append(x1)
                    i += 1
                contour = np.asarray(ncontour)

                # if len(contour) > contourGap * 4:
                #     contour = contour[0::contourGap]
                newcontours.append((contour, actualarea))
            # sort by area size
            newcontours = sorted(newcontours, key=lambda x: len(x[0]), reverse=True)
            # print("Contour: ", [ len(c) for c, a in newcontours])
            found_contours[label_name] = newcontours

    def saveMask(self, im, dst_dir, imname, \
        maskperimage=-1, contourGap=6, contourArea=None, contourThresh=None,\
        skipUnmask=True, radius = 5, withoutMask=False, \
        style="labelme", **kwargs):
        """@Note setPrediciton must be called before this method invoked
        save mask / keypoints

        Arguments:
            im(torch.Tensor): image object processed by Dataset/Model, 
                it should be the same with the original image no matter how it is processed (preprocess & postprocess).
                **Note, if the Dataset/Model didnot reveal the image, you'd better pass a original Image object**
            imname(str): image pure name withou suffix
            style(str): only support `labelme` currently
            maskperimage(int): if -1, then no mask limit for each image, otherwize each class will only get certain masks
        """
        os.makedirs(osp.join(dst_dir, "Annotation"), exist_ok=True)
        im = self.convertImg(im)
        
        if contourThresh is None:
            contourThresh = [127, 255, 0]

        # print(found_contours)
        shapes = []
        ## add mask type as polygon
        if not withoutMask:
            found_contours = { c: None for c in self._captions }
            ## TODO contour will change orders
            self._maskToContour(found_contours, contourGap, contourThresh, contourArea)
            for label_name, contours in found_contours.items():
                mc = 0
                if contours is None:
                    continue
                for contour, areasize in contours:
                    if maskperimage > 0 and mc >= maskperimage:
                        break
                    mc += 1
                    shapes.append({"label": label_name, 
                        "points": contour.tolist(),       
                        "group_id": None,
                        "shape_type": "polygon",
                        "flags": {}
                    })
        if self.keypoints is not None:
            ## TODO how to map keypoints to label better?
            kpNum = self.keypoints.shape[0]
            mc = 0
            for ci in range(kpNum):
                kps = self.keypoints[ci]
                if maskperimage > 0 and mc >= maskperimage:
                    break

                label_name = self._captions[self.labels[ci]]
                for i, kp in enumerate(kps):
                    if self.kpcaptions is None:
                        ln = f"{label_name}-{i}"
                    else:
                        ln = self.kpcaptions[i]
                    # print(ln, kp)
                    shapes.append({"label": ln, 
                        "points": [ [ float(kp[0]), float(kp[1]) ] ],
                        "group_id": None,
                        "shape_type": "point",
                        "flags": {}
                    })
        if len(shapes) == 0 and skipUnmask:
            return PredictionViewer.Save_SkippedNoMask
        # print(f"mask will be saved: {len(shapes)}")

        if imname.endswith(self.img_suffix):
            imname = imname[:imname.rfind(self.img_suffix)]
        # imname should withoud suffix
        self._save(im, osp.join(dst_dir, "Annotation"), imname, \
            shapes, im.shape)
        return PredictionViewer.Save_Success

    def saveViz(self, im, dst_dir, imname, mask_threshold = 127, radius=3, **kwargs):
        """
        Arguments:
            process_func (function):
        """
        if imname.endswith(self.img_suffix):
            imfilename = imname # full name
            imname = imname[:imname.rfind(self.img_suffix)]
        else:
            imfilename = f"{imname}{self.img_suffix}"

        ## save visualized image
        os.makedirs(osp.join(dst_dir, "viz"), exist_ok=True)
        annIm = self.show(im, mask_threshold=mask_threshold, radius=radius)
        annIm.save(osp.join(dst_dir, "viz", imfilename))


import torch.distributed as dist
from .imageutils import ImagePreprocessor, CvImageReader

def warmup_lr_scheduler(optimizer, warmup_iters, warmup_factor):

    def f(x):
        if x >= warmup_iters:
            return 1
        alpha = float(x) / warmup_iters
        return warmup_factor * (1 - alpha) + alpha

    return torch.optim.lr_scheduler.LambdaLR(optimizer, f)

class F1ScoreCalcor():
    """Binary Label F1-score calculator
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.TP = 0
        self.FP = 0
        self.FN = 0

    def compare(self, preds, targets):
        """
        Arguments:
            preds(Tensor): (n x 1)
            targets(Tensor): (n x 1)
        """
        self.TP += np.sum(preds & (targets == 1))
        self.FP += np.sum(preds & (targets == 0))
        self.FN += np.sum(~preds & (targets == 1))

    def export(self, printit=True):
        TP = self.TP
        FP = self.FP
        FN = self.FN
        precision = TP / (TP+FP)
        recall = TP / (TP + FN)
        f1score = 2 * precision * recall / (precision + recall)
        if printit:
            print(f"TP {self.TP}, FP {self.FP}, FN {self.FN}, Precision: {precision:.6f}, Recall: {recall}, F1-score: {f1score}")
        return precision, recall, f1score

class PearsonCorrelation(torch.nn.Module):


    def forward(self,tensor_1,tensor_2):
        x = tensor_1
        y = tensor_2

        vx = x - torch.mean(x)
        vy = y - torch.mean(y)

        cost = torch.sum(vx * vy) / (torch.sqrt(torch.sum(vx ** 2)) * torch.sqrt(torch.sum(vy ** 2)))
        return cost

class PearsonClassfication():
    """
    Arguments:
        base (torch.Tensor)
        converter( Callable ): which is used to convert imgs into torch.Tensor if not None
    """
    def __init__(self, base, converter=None):
        self.pc = PearsonCorrelation()
        self.converter = converter
        if not isinstance(base, torch.Tensor) and self.converter is not None:
            base = self.converter(base)
        self.base = base
        assert isinstance(self.base, torch.Tensor)

    def compare(self, imglists, thresh, names = None):
        """
        Arguments:
            imglists(list(torch.Tensor) Iterable) it could be a list which contains torch.Tensor instances,
                or it could be just torch.utils.data.Dataset which is iterable for getting images,
                if elemetns in list is not instance of torch.Tensor and converter is not None,
                then it will be converted first
            names(list(str) | None)
        """
        similar = []
        unsimilar = []
        corr = []
        self.pc.eval()
        print("Length of list to be checked: ", len(imglists))
        with torch.no_grad():
            for idx, img in enumerate(imglists):
                if type(img) is tuple:
                    img = img[0]
                if not isinstance(img, torch.Tensor) and self.converter is not None:
                    img = self.converter(img)
                assert isinstance(img, torch.Tensor)

                c = self.pc(self.base, img)
                corr.append(c)
                name = names[idx] if names is not None else ""
                if c > thresh:
                    similar.append((img, name))
                else:
                    unsimilar.append((img, name))
        return similar, unsimilar, corr


def draw_CAM(model, img_path, save_path, transform=None, visual_heatmap=False):
    '''
    绘制 Class Activation Map
    :param model: 加载好权重的Pytorch model
    :param img_path: 测试图片路径
    :param save_path: CAM结果保存路径
    :param transform: 输入图像预处理方法
    :param visual_heatmap: 是否可视化原始heatmap（调用matplotlib）
    :return:
    '''
    # 图像加载&预处理
    img = Image.open(img_path).convert('RGB')
    if transform:
        img = transform(img)
    img = img.unsqueeze(0)
 
    # 获取模型输出的feature/score
    model.eval()
    features = model.features(img)
    output = model.classifier(features)
 
    # 为了能读取到中间梯度定义的辅助函数
    def extract(g):
        global features_grad
        features_grad = g
 
    # 预测得分最高的那一类对应的输出score
    pred = torch.argmax(output).item()
    pred_class = output[:, pred]
 
    features.register_hook(extract)
    pred_class.backward() # 计算梯度
 
    grads = features_grad   # 获取梯度
 
    pooled_grads = torch.nn.functional.adaptive_avg_pool2d(grads, (1, 1))
 
    # 此处batch size默认为1，所以去掉了第0维（batch size维）
    pooled_grads = pooled_grads[0]
    features = features[0]
    # 512是最后一层feature的通道数
    for i in range(512):
        features[i, ...] *= pooled_grads[i, ...]
 
    # 以下部分同Keras版实现
    heatmap = features.detach().numpy()
    heatmap = np.mean(heatmap, axis=0)
 
    heatmap = np.maximum(heatmap, 0)
    heatmap /= np.max(heatmap)
 
    # 可视化原始热力图
    if visual_heatmap:
        plt.matshow(heatmap)
        plt.show()
 
    img = cv.imread(img_path)  # 用cv2加载原始图像
    heatmap = cv.resize(heatmap, (img.shape[1], img.shape[0]))  # 将热力图的大小调整为与原始图像相同
    heatmap = np.uint8(255 * heatmap)  # 将热力图转换为RGB格式
    heatmap = cv.applyColorMap(heatmap, cv.COLORMAP_JET)  # 将热力图应用于原始图像
    superimposed_img = heatmap * 0.4 + img  # 这里的0.4是热力图强度因子
    cv.imwrite(save_path, superimposed_img)  # 将图像保存到硬盘


def getSGDOptimizer(model, lr=0.005, weight_decay=0.00002, step_size=6, **kwargs):
    params = [p for p in model.parameters() if p.requires_grad]
    # for name, param in model.named_parameters():
    #      if param.requires_grad:
    #         print(name)

    optimizer = torch.optim.SGD(params, lr=lr,
                                momentum=0.9, weight_decay=weight_decay)
    # and a learning rate scheduler
    lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer,
                                                step_size=step_size,
                                                gamma=kwargs.get("gamma", 0.2))
    return optimizer, lr_scheduler


#!/usr/env/bin python3
# -*- coding: utf-8 -*-

__version__ = '0.0.2'
from bffacilities.utils import createLogger, initGetText
from bffacilities._frame import Frame, common_entry
from pathlib import Path
from argparse import ArgumentParser
import os
from colorama import Fore, Style

tr = initGetText("gitrepo")
logger = createLogger("gitrepo.log", stream=True, savefile=False)

class PaperUtil(Frame):
    def __init__(self):
        self.config = {}
        # self.papers = []

    def readConfig(self, file):
        p = Path(".papers")
        p.mkdir(parents=True, exist_ok=True)

    def initArgs(self):
        """ @return ArgumentParser """
        parser = ArgumentParser(description=tr("Used for check downloaded papers and whether a paper is downloaded."))

        action = parser.add_mutually_exclusive_group()
        action.add_argument("-c", "--check", help=tr("Check whether the papers is downloaded ."))
        action.add_argument("-l", "--list", action="store_true",
                            help=tr("List existing papers under <{}>.").format(os.getcwd()))
        action.add_argument("-o", "--open", help=tr("Opening papers"))
        action.add_argument("-a", "--add", help=tr("Add a paper that needs to be download"))
        action.add_argument("-L", "--list-tmp", action="store_true", help=tr("elete all tmp record for papers that need to be download"))
        action.add_argument("-D", "--delete-tmp", action="store_true", help=tr("Delete all tmp record for paper already downloaded"))

        parser.add_argument("-V", "--version", 
            help=tr("Print this script version"), action="version", version=f'paperUtil {__version__}')
        self.parser = parser

    def parseArgs(self, argv):
        args = self.parser.parse_args(argv)
        if args.check:
            check = args.check
            self._check_paper(check, False)
        elif args.open:
            self._open_paper(args.open)
        elif args.add:
            self._add(args.add)
        elif args.list_tmp:
            try:
                with open(".papers/record.txt", "r", encoding="utf-8") as f:
                    for l in f:
                        print(l.strip())
            except:
                pass
        elif args.delete_tmp:
            with open(".papers/record.txt", "w", encoding="utf-8") as f:
                f.write("")
        else:
            # default action
            self._list_papers(os.getcwd())

    def _add(self, a):
        possible = self._find_possible_exist(a)
        confirm = True
        if possible:
            uinput = input(f"Still Add {a}: \n")
            confirm = (uinput.strip().upper() == 'Y')
            print(uinput, confirm)
        if confirm:
            with open(".papers/record.txt", "a+", encoding="utf-8") as f:
                f.write(a + "\n")

    def _list_papers(self, dir, printable=True):
        path = Path(dir)
        dirs = [x for x in path.iterdir() if x.is_dir() and x.name.isdigit()]
        if printable: print('current', path)
        paperCount = 0
        self.papers = {} 
        for p in dirs:
            pdirs = [x.name.strip() for x in p.iterdir() if x.is_file()]
            self.papers[p.name] = pdirs
            paperCount += len(pdirs)
            if not printable:
                continue
            print(f"folder: {p.name} ({len(pdirs)})")
            for d in pdirs:
                # byte = d.encode()
                try:
                    print('- ', d)
                except:
                    print(f'{Fore.YELLOW}- {d.encode()}{Style.RESET_ALL}')
            print('--------------------\n')
        if printable:
            print(f'Summary: {paperCount} papers')

    def _find_possible_exist(self, name: str):
        self._list_papers(os.getcwd(), False)
        name = name.strip()
        found = False
        for pdir, files in self.papers.items():
            for f in files:
                upperF = f.upper()
                if upperF.find(name.upper()) >= 0:
                    print(f"{Fore.YELLOW}[Possible] {f}{Style.RESET_ALL}")
                    found = True
        return found

    def _check_paper(self, name: str, strict=True):
        self._list_papers(os.getcwd(), False)
        name = name.strip()
        if strict:
            for pdir, files in self.papers.items():
                if name in files:
                    print(f"{Fore.GREEN}[Found] {name}{Style.RESET_ALL}")
                    return pdir
        else:
            found = self._find_possible_exist(name)
            if found:
                print(f"\n{Fore.GREEN}[Found] {name}{Style.RESET_ALL}")
            else:
                print(f"\n{Fore.RED}[Not Found] {name}{Style.RESET_ALL}")
        return None

    def _open_paper(self, name):
        import subprocess, platform
        curDir = os.getcwd()
        fdir = self._check_paper(name)
        if fdir is None:
            return
        try:
            if platform.system() == 'Darwin':       # macOS
                file = f"{curDir}/{fdir}/'{name}'"
                subprocess.call(('open', file))
            elif platform.system() == 'Windows':    # Windows
                file = f"{curDir}\\{fdir}\\{name}"
                os.startfile(file)
                # subprocess.run(('open', file), check=True)
            else:                                   # linux variants
                file = f"{curDir}/{fdir}/'{name}'"
                subprocess.call(('xdg-open', file))
        except Exception as e:
            logger.warning(e)
main = common_entry(PaperUtil)


"""some basic configurations
"""
from pathlib import Path
import os
import os.path as osp

_BASE_DIR = Path.home() / ".shells" # Basic Path
_CONFIG_DIR = _BASE_DIR / "configs"  # Default Config Path

_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

BFF_ROOT_PATH = Path(osp.abspath(osp.dirname(__file__)))
BFF_OTHER_PATH = BFF_ROOT_PATH / "other"
BFF_TEMPLATE_PATH = BFF_ROOT_PATH / "template"